"""Fortran statments and expressions supported by current KGen

Module content
---------------
"""

# kgen_search.py

from fortlab.kgutils import show_tree, logger
from fortlab.resolver.kgparse import KGGenType
import fortlab.resolver.Fortran2003 as Fortran2003
#from typedecl_statements import TypeDeclarationStatement, TypeStmt, Procedure # TEEMP
from fortlab.resolver.typedecl_statements import TypeDeclarationStatement, TypeStmt
from fortlab.resolver.block_statements import Type, TypeDecl, Function, Subroutine, Interface, execution_part, Associate
from fortlab.resolver.statements import External, Common, SpecificBinding, Enumerator
from collections import OrderedDict

#import logging
#logger = logging.getLogger('kgen')

res_default = [ TypeDeclarationStatement ]
#res_external = [ External, Procedure ] # TEMP
res_external = [ External ]
res_typedecl = [ TypeDeclarationStatement ]
res_typestmt = [ TypeStmt ]
res_derivedtype = [ Type, TypeDecl ] 
res_associate = [ Associate ] 
res_kind = [ TypeDeclarationStatement ] + res_derivedtype
res_typespec = [ TypeDeclarationStatement ] + res_derivedtype
res_value = ([ TypeDeclarationStatement, Function, Interface, Enumerator ] +
             res_external + res_associate)
res_subroutine = [ Subroutine, Interface ] + res_external
res_function = [ Function, Interface ] + res_external
res_subprogram = [ Subroutine, Function, Interface ] + res_external
res_common = [ Common ]
res_ptr_object = [ SpecificBinding, TypeDeclarationStatement ]
res_target = res_subprogram + res_typedecl
res_anything = res_typespec + res_subprogram + [ SpecificBinding, Common, Type, TypeDecl ]

###############################################################################
################################### COMMON ####################################
###############################################################################

class SearchException(Exception):
    pass

def f2003_search_unknowns(stmt, node, config, resolvers=None, gentype=None):
    """Identify unknowns whose declaration statement will be searched by KGen.

    Parameters
    ----------
    stmt : F2PY parser statement object
        Specify a statement object to be searched
    node : F2PY Fortran2003 parser object
        Specify an expression object to be searched
    resolvers : A list of statement classes for resolver
        Limits the classes of resolver
    gentype : Type of state data (IN or OUT)
        Specify the type of state data

    Returns
    -------
    None

    See also
    --------
    get_name_or_defer
    get_name
    defer
    defer_names
    """

    if node is None: return

    # save in unknowns dict in stmt
    if not hasattr(stmt, 'unknowns'):
        stmt.unknowns = OrderedDict()

    # skip searching if specified
    if ( hasattr(node, 'skip_search') and node.skip_search ) or \
        ( hasattr(node, 'parent') and hasattr(node.parent, 'skip_search') and node.parent.skip_search ):
        return

    clsname = node.__class__.__name__

    if clsname=='Name':
        get_name(stmt, node, resolvers, config, gentype=gentype)
        return

    itemclsname = None
    try:
        if clsname.endswith('_List'):
            _clsname = clsname[:-5]
            for item in node.items:
                if item is None: continue
                itemclsname = item.__class__.__name__
                if itemclsname=='Name':
                    get_name(stmt, item, resolvers, config, gentype=gentype)
                else:
                    exec('search_%s(stmt, item, config, gentype=gentype)' % itemclsname)
        elif clsname.startswith('End_'):
            pass
        else:
            exec('search_%s(stmt, node, config, gentype=gentype)' % clsname)
    except Exception as e:
        errname = clsname
        if itemclsname:
            errname = itemclsname
        errmsg = "Error: Fortran specification of %s is not supported yet."%errname

        logger.exception(errmsg)

        if config["search"]['promote_exception']:
            raise
        else:
            print('')
            print(errmsg)
            print('')
            print("'kgen.log' in output folder contains detail information of this error.")
            print("If you send the log file to 'kgen@ucar.edu', that could be very")
            print("helpful for us to support this Fortran spec. in future KGEN version.")
            print('')
            import sys
            sys.exit(-1)

def get_name_or_defer(stmt, node, resolvers, config, defer=True, gentype=None):
    """Select a name to be searched, or defer to lower level of nodes in AST.

    Parameters
    ----------
    stmt : F2PY parser statement object
        Specify a statement object to be searched
    node : F2PY Fortran2003 parser object
        Specify an expression object to be searched
    resolvers : A list of statement classes for resolver
        Limits the classes of resolver
    defer : bool
        check if to search lower level of nodes in AST.
    gentype : Type of state data (IN or OUT)
        Specify the type of state data

    Returns
    -------
    None

    See also
    --------
    f2003_search_unknowns
    get_name
    defer
    defer_names
    """

    from fortlab.kgutils import KGName, pack_innamepath, match_namepath
    from fortlab.resolver.kgparse import ResState
    from fortlab.resolver.kgintrinsics import Intrinsic_Procedures
    from fortlab.resolver.base_classes import is_except

    if node is None: return

    # uncomment below line for debug
    #print node.__class__, str(node)

    if isinstance(node, Fortran2003.Name):

        # skip if intrinsic
        if node.string.lower() in Intrinsic_Procedures:
            excepts = config["search"]['except']

            if  config["search"]['skip_intrinsic'] and not is_except(node, stmt, excepts):
                if hasattr(node, 'parent') and not isinstance(node.parent, Fortran2003.Part_Ref) and \
                    not (isinstance(node.parent, Fortran2003.Function_Reference) and node.string.lower()=='null') and \
                    not (isinstance(node.parent, Fortran2003.Specific_Binding) and node.string.lower()=='null'):
                    logger.debug('Intrinsic procedure name of "%s" is used for name resolution'% \
                        (node.string.lower()))
                    logger.debug('\tnear "%s"'% stmt.item.line)
                    logger.debug('\tin %s'% stmt.reader.id)
                else:
                    #if node.string.lower()!='null':
                    #    logger.debug('Intrinsic procedure name of "%s" is skipped from name resolution'% \
                    #        (node.string.lower()))
                    #logger.debug('\tnear "%s"'% stmt.item.line)
                    #logger.debug('\tin %s'% stmt.reader.id)
                    return
    
            elif not config["search"]['skip_intrinsic'] and is_except(node, stmt, excepts): 
                if hasattr(node, 'parent') and not isinstance(node.parent, Fortran2003.Part_Ref) and \
                    not (isinstance(node.parent, Fortran2003.Function_Reference) and node.string.lower()=='null') and \
                    not (isinstance(node.parent, Fortran2003.Specific_Binding) and node.string.lower()=='null'):
                    #logger.debug('Intrinsic procedure name of "%s" is NOT skipped from name resolution'% \
                    #    (node.string.lower()))
                    #logger.debug('\tnear "%s"'% stmt.item.line)
                    #logger.debug('\tin %s'% stmt.reader.id)
                    pass
                else:
                    if node.string.lower()!='null':
                        logger.debug('Intrinsic procedure name of "%s" is skipped from name resolution'% \
                            (node.string.lower()))
                    logger.debug('\tnear "%s"'% stmt.item.line)
                    logger.debug('\tin %s'% stmt.reader.id)
                    return

        # skip if excluded
        #if config.exclude.has_key('namepath') and stmt.__class__ in execution_part:
        if isinstance(config, list): import pdb; pdb.set_trace()
        if 'namepath' in config["exclude"]:
            for pattern, actions in config["exclude"]['namepath'].items():
                name = node.string.lower()
                namepath = pack_innamepath(stmt, name) 
                #logger.debug('%s and %s are being checked for exclusion'%(pattern, namepath))
                if match_namepath(pattern, namepath):
                    #logger.debug('%s and %s are mathched for exclusion'%(pattern, namepath))
                    if not hasattr(stmt, 'exclude_names'): stmt.exclude_names = OrderedDict()
                    if name in stmt.exclude_names:
                        stmt.exclude_names[name].extend(actions)
                    else:
                        stmt.exclude_names[name] = actions
                    node.skip_search = True
                    if hasattr(node, 'parent'): node.parent.skip_search = True
                    return

        ukey = KGName(pack_innamepath(stmt, node.string.lower()), node=node, stmt=stmt)

        if gentype is None: gentype = KGGenType.STATE_IN

        if resolvers is None:
            stmt.unknowns[ukey] = ResState(gentype, ukey, stmt, res_default)
        else:
            stmt.unknowns[ukey] = ResState(gentype, ukey, stmt, resolvers)
        logger.debug('%s is saved as unknown' % node.string.lower())

    elif defer:
        f2003_search_unknowns(stmt, node, config, resolvers, gentype=gentype)

def get_name(stmt, node, resolvers, config, gentype=None):
    get_name_or_defer(stmt, node, resolvers, config, defer=False, gentype=gentype)

def defer(stmt, node, config, gentype=None):
    if isinstance(node, Fortran2003.Name):
        raise SearchException('%s can not be Name class' % str(node))
    f2003_search_unknowns(stmt, node, config, gentype=gentype)


def defer_items(stmt, node, config, gentype=None):
    if hasattr(node, 'items'):
        for item in node.items:
            if isinstance(item, Fortran2003.Name):
                raise SearchException('%s can not be Name class' % str(item))
            f2003_search_unknowns(stmt, item, config, gentype=gentype)

###############################################################################
################################### SEARCH ####################################
###############################################################################

def search_Type_Declaration_Stmt(stmt, node, config, gentype=None):  
    """ Identifying a name in Type_Declaration_Stmt node"""

    from fortlab.kgutils import pack_innamepath, match_namepath

    # collect excluded names
    if 'namepath' in config["exclude"]:
        for pattern, actions in config["exclude"]['namepath'].items():
            decls = []
            if isinstance(node.items[2], Fortran2003.Entity_Decl):
                decls.append(node.items[2].items[0].string.lower())
            elif isinstance(node.items[2], Fortran2003.Entity_Decl_List):
                for item in node.items[2].items:
                    decls.append(item.items[0].string.lower())
            for decl in decls:
                namepath = pack_innamepath(stmt, decl) 
                if match_namepath(pattern, namepath):
                    if not hasattr(stmt, 'exclude_names'): stmt.exclude_names = OrderedDict()
                    if decl in stmt.exclude_names:
                        stmt.exclude_names[decl].extend(actions)
                    else:
                        stmt.exclude_names[decl] = actions

    defer_items(stmt, node, config)

def search_Intrinsic_Type_Spec(stmt, node, config, gentype=None): 
    """ Identifying a name in Intrinsic_Type_Spec node"""
    defer(stmt, node.items[1], config)

def search_Kind_Selector(stmt, node, config, gentype=None): 
    """ Identifying a name in Kind_Selector node"""
    get_name_or_defer(stmt, node.items[1], res_kind, config)

def search_Entity_Decl(stmt, node, config, gentype=None): 
    """ Identifying a name in Entity_Decl node"""
    defer(stmt, node.items[1], config)
    get_name_or_defer(stmt, node.items[2], res_value, config)
    get_name_or_defer(stmt, node.items[3], res_value, config) 

def search_Explicit_Shape_Spec(stmt, node, config, gentype=None): 
    """ Identifying a name in Explicit_Shape_Spec node"""
    get_name_or_defer(stmt, node.items[0], res_value, config)
    get_name_or_defer(stmt, node.items[1], res_value, config)

def search_Dimension_Attr_Spec(stmt, node, config, gentype=None): 
    """ Identifying a name in Dimension_Attr_Spec node"""
    defer(stmt, node.items[1], config)

def search_Add_Operand(stmt, node, config, gentype=None): 
    """ Identifying a name in Add_Operand node"""
    get_name_or_defer(stmt, node.items[0], res_value, config)
    get_name_or_defer(stmt, node.items[2], res_value, config)

def search_Mult_Operand(stmt, node, config, gentype=None): 
    """ Identifying a name in Mult_Operand node"""
    get_name_or_defer(stmt, node.items[0], res_value, config)
    get_name_or_defer(stmt, node.items[2], res_value, config)

def search_Attr_Spec(stmt, node, config, gentype=None): 
    """ Identifying a name in Attr_Spec node"""
    defer_items(stmt, node, config)

def search_Initialization(stmt, node, config, gentype=None): 
    """ Identifying a name in Initialization node"""
    get_name_or_defer(stmt, node.items[1], res_value, config)

def search_Part_Ref(stmt, node, config, gentype=None): 
    """ Identifying a name in Part_Ref node"""
    get_name_or_defer(stmt, node.items[0], res_value, config, gentype=gentype) 
    get_name_or_defer(stmt, node.items[1], res_value, config) 

def search_Structure_Constructor_2(stmt, node, config, gentype=None): 
    """ Identifying a name in Structure_Constructor_2 node"""
    get_name_or_defer(stmt, node.items[1], res_value, config)

def search_Int_Literal_Constant(stmt, node, config, gentype=None): 
    """ Identifying a name in Int_Literal_Constant node"""
    if node.items[1]:
        get_name_or_defer(stmt, Fortran2003.Name(node.items[1]), res_typedecl, config)

def search_Signed_Int_Literal_Constant(stmt, node, config, gentype=None): 
    """ Identifying a name in Signed_Int_Literal_Constant node"""
    if node.items[1]:
        get_name_or_defer(stmt, Fortran2003.Name(node.items[1]), res_typedecl, config)

def search_Real_Literal_Constant(stmt, node, config, gentype=None): 
    """ Identifying a name in Real_Literal_Constant node"""
    if node.items[1]:
        get_name_or_defer(stmt, Fortran2003.Name(node.items[1]), res_typedecl, config)

def search_Signed_Real_Literal_Constant(stmt, node, config, gentype=None): 
    """ Identifying a name in Signed_Real_Literal_Constant node"""
    if node.items[1]:
        get_name_or_defer(stmt, Fortran2003.Name(node.items[1]), res_typedecl, config)

def search_Subroutine_Stmt(stmt, node, config, gentype=None): 
    """ Identifying a name in Subroutine_Stmt node"""
    get_name_or_defer(stmt, node.items[2], res_typedecl, config) # dummy args
    get_name_or_defer(stmt, node.items[3], res_typedecl, config) # postfix

def search_Comment(stmt, node, config, gentype=None): 
    """ Identifying a name in Comment node"""
    if hasattr(stmt, 'write_state'):
        for var in stmt.write_state:
            f2003obj = Fortran2003.Variable(var)
            get_name_or_defer(stmt, f2003obj, res_typedecl, config)

def search_Nonlabel_Do_Stmt(stmt, node, config, gentype=None): 
    """ Identifying a name in Nonlabel_Do_Stmt node"""
    if len(node.items)==3:
        defer(stmt, node.items[2], config)
    elif len(node.items)==2:
        if isinstance(node.items[0], str):
            defer(stmt, node.items[1], config)

def search_Loop_Control(stmt, node, config, gentype=None): 
    """ Identifying a name in Loop_Control node"""
    if len(node.items)==1:
        get_name_or_defer(stmt, node.items[0], res_value, config)
    else:
        get_name_or_defer(stmt, node.items[0], res_typedecl, config, gentype=KGGenType.STATE_OUT)
        if isinstance(node.items[1], list):
            for item in node.items[1]:
                get_name_or_defer(stmt, item, res_value, config)
        else:
            get_name_or_defer(stmt, node.items[1], res_value, config)

def search_Assignment_Stmt(stmt, node, config, gentype=None): 
    """ Identifying a name in Assignment_Stmt node"""
    get_name_or_defer(stmt, node.items[0], res_value, config, gentype=KGGenType.STATE_OUT)
    get_name_or_defer(stmt, node.items[2], res_value, config)

def search_Level_2_Expr(stmt, node, config, gentype=None): 
    """ Identifying a name in Level_2_Expr node"""
    get_name_or_defer(stmt, node.items[0], res_value, config)
    get_name_or_defer(stmt, node.items[2], res_value, config)

def search_Parenthesis(stmt, node, config, gentype=None): 
    """ Identifying a name in Parenthesis node"""
    get_name_or_defer(stmt, node.items[1], res_value, config, gentype=gentype)

def search_str(stmt, string, config, gentype=None):
    pass

def search_Function_Stmt(stmt, node, config, gentype=None): 
    """ Identifying a name in Function_Stmt node"""
    get_name_or_defer(stmt, node.items[0], res_derivedtype, config ) # prefix
    get_name_or_defer(stmt, node.items[2], res_typedecl, config) # dummy args
    get_name_or_defer(stmt, node.items[3], res_typedecl, config)

def search_Assumed_Shape_Spec(stmt, node, config, gentype=None): 
    """ Identifying a name in Assumed_Shape_Spec node"""
    get_name_or_defer(stmt, node.items[0], res_value, config)
    get_name_or_defer(stmt, node.items[1], res_value, config)

def search_Allocate_Stmt(stmt, node, config, gentype=None): 
    """ Identifying a name in Allocate_Stmt node"""
    get_name_or_defer(stmt, node.items[0], res_typespec, config)
    get_name_or_defer(stmt, node.items[1], res_typedecl, config)
    defer(stmt, node.items[2], config)

def search_Allocation(stmt, node, config, gentype=None): 
    """ Identifying a name in Allocation node"""
    get_name_or_defer(stmt, node.items[0], res_typedecl, config)
    get_name_or_defer(stmt, node.items[1], res_value, config)
    #if len(node.items)>1:
    #    defer_items(stmt, node.items[1:])

def search_Allocate_Shape_Spec(stmt, node, config, gentype=None): 
    """ Identifying a name in Allocate_Shape_Spec node"""
    if node.items:
        for item in node.items:
            get_name_or_defer(stmt, item, res_value, config)

def search_Use_Stmt(stmt, node, config, gentype=None): 
    """ Identifying a name in Use_Stmt node"""
    pass

def search_If_Then_Stmt(stmt, node, config, gentype=None): 
    """ Identifying a name in If_Then_Stmt node"""
    get_name_or_defer(stmt, node.items[0], res_value, config)

def search_Level_4_Expr(stmt, node, config, gentype=None): 
    """ Identifying a name in Level_4_Expr node"""
    get_name_or_defer(stmt, node.items[0], res_value, config)
    get_name_or_defer(stmt, node.items[2], res_value, config)

def search_If_Stmt(stmt, node, config, gentype=None): 
    """ Identifying a name in If_Stmt node"""
    get_name_or_defer(stmt, node.items[0], res_value, config)
    get_name_or_defer(stmt, node.items[1], res_value, config)

def search_Else_If_Stmt(stmt, node, config, gentype=None): 
    """ Identifying a name in Else_If_Stmt node"""
    get_name_or_defer(stmt, node.items[0], res_value, config)
    get_name_or_defer(stmt, node.items[1], res_value, config)

def search_Else_Stmt(stmt, node, config, gentype=None): 
    """ Identifying a name in Else_Stmt node"""
    pass

def search_Level_2_Unary_Expr(stmt, node, config, gentype=None): 
    """ Identifying a name in Level_2_Unary_Expr node"""
    get_name_or_defer(stmt, node.items[1], res_value, config)


def search_Label_Do_Stmt(stmt, node, config, gentype=None): 
    """ Identifying a name in Label_Do_Stmt node"""
    defer(stmt, node.items[2], config)

def search_Array_Constructor(stmt, node, config, gentype=None): 
    """ Identifying a name in Array_Constructor node"""
    get_name_or_defer(stmt, node.items[1], res_value, config) 

def search_Array_Section(stmt, node, config, gentype=None): 
    """ Identifying a name in Array_Section node"""
    get_name_or_defer(stmt, node.items[0], res_value, config)
    defer(stmt, node.items[1], config)

def search_Substring_Range(stmt, node, config, gentype=None): 
    """ Identifying a name in Substring_Range node"""
    get_name_or_defer(stmt, node.items[0], res_value, config)
    get_name_or_defer(stmt, node.items[1], res_value, config)

def search_Select_Case_Stmt(stmt, node, config, gentype=None): 
    """ Identifying a name in Select_Case_Stmt node"""
    get_name_or_defer(stmt, node.items[0], res_value, config)

def search_Case_Stmt(stmt, node, config, gentype=None): 
    """ Identifying a name in Case_Stmt node"""
    get_name_or_defer(stmt, node.items[0], res_value, config)

def search_Case_Selector(stmt, node, config, gentype=None): 
    """ Identifying a name in Case_Selector node"""
    get_name_or_defer(stmt, node.items[0], res_value, config)

def search_Call_Stmt(stmt, node, config, gentype=None): 
    """ Identifying a name in Call_Stmt node"""
    get_name_or_defer(stmt, node.items[0], res_subroutine, config)
    #if isinstance(node.items[1], Fortran2003.Name):
    #    get_name_or_defer(stmt, node.items[1], res_value)
    #else:
    #    defer(stmt, node.items[1])
    get_name_or_defer(stmt, node.items[1], res_value, config)

def search_Char_Literal_Constant(stmt, node, config, gentype=None): 
    """ Identifying a name in Char_Literal_Constant node"""
    if node.items[1]:
        get_name_or_defer(stmt, Fortran2003.Name(node.items[1]), res_typedecl, config)
    #get_name_or_defer(stmt, node.items[0], res_typedecl)

def search_Length_Selector(stmt, node, config, gentype=None): 
    """ Identifying a name in Length_Selector node"""
    for item in node.items:
        get_name_or_defer(stmt, item, res_value, config)

def search_Type_Param_Value(stmt, node, config, gentype=None): 
    """ Identifying a name in Type_Param_Value node"""
    # NOTE: need to verify its content structure
    if node.item:
        get_name_or_defer(stmt, node.item, res_value, config)

def search_Write_Stmt(stmt, node, config, gentype=None): 
    """ Identifying a name in Write_Stmt node"""
    get_name_or_defer(stmt, node.items[0], res_value, config)
    get_name_or_defer(stmt, node.items[1], res_value, config)

def search_Read_Stmt(stmt, node, config, gentype=None): 
    """ Identifying a name in Read_Stmt node"""
    get_name_or_defer(stmt, node.items[0], res_value, config)
    get_name_or_defer(stmt, node.items[1], res_value, config)
    get_name_or_defer(stmt, node.items[2], res_value, config)

def search_Io_Control_Spec(stmt, node, config, gentype=None): 
    """ Identifying a name in Io_Control_Spec node"""
    get_name_or_defer(stmt, node.items[1], res_value, config)

def search_Stop_Stmt(stmt, node, config, gentype=None): 
    """ Identifying a name in Stop_Stmt node"""
    pass

def search_Contains_Stmt(stmt, node, config, gentype=None): 
    """ Identifying a name in Contains_Stmt node"""
    pass

def search_Subscript_Triplet(stmt, node, config, gentype=None): 
    """ Identifying a name in Subscript_Triplet node"""
    get_name_or_defer(stmt, node.items[0], res_value, config) 
    get_name_or_defer(stmt, node.items[1], res_value, config) 
    get_name_or_defer(stmt, node.items[2], res_value, config) 

def search_Interface_Stmt(stmt, node, config, gentype=None):
    """ Identifying a name in Interface_Stmt node"""
    pass

def search_Procedure_Stmt(stmt, node, config, gentype=None):
    """ Identifying a name in Procedure_Stmt node"""
    get_name_or_defer(stmt, node.items[0], res_subprogram, config)

def search_Prefix(stmt, node, config, gentype=None):
    """ Identifying a name in Prefix node"""
    for item in node.items:
        get_name_or_defer(stmt, node.items[0], res_anything, config)

def search_Prefix_Spec(stmt, node, config, gentype=None):
    """ Identifying a name in Prefix_Spec node"""
    if node.item or hasattr(node, 'items'):
        raise ProgramException('Unexpected item or items attr')

def search_Logical_Literal_Constant(stmt, node, config, gentype=None):
    """ Identifying a name in Logical_Literal_Constant node"""
    if node.items[1]:
        get_name_or_defer(stmt, Fortran2003.Name(node.items[1]), res_typedecl, config)
    #get_name_or_defer(stmt, node.items[1], res_typedecl)

def search_Access_Spec(stmt, node, config, gentype=None):
    """ Identifying a name in Access_Spec node"""
    pass

def search_And_Operand(stmt, node, config, gentype=None):
    """ Identifying a name in And_Operand node"""
    get_name_or_defer(stmt, node.items[1], res_value, config)

def search_Equiv_Operand(stmt, node, config, gentype=None):
    """ Identifying a name in Equiv_Operand node"""
    get_name_or_defer(stmt, node.items[0], res_value, config)
    get_name_or_defer(stmt, node.items[2], res_value, config)

def search_Or_Operand(stmt, node, config, gentype=None):
    """ Identifying a name in Or_Operand node"""
    get_name_or_defer(stmt, node.items[0], res_value, config)
    get_name_or_defer(stmt, node.items[2], res_value, config)

def search_Where_Construct_Stmt(stmt, node, config, gentype=None):
    """ Identifying a name in Where_Construct_Stmt node"""
    get_name_or_defer(stmt, node.items[0], res_value, config)

def search_Elsewhere_Stmt(stmt, node, config, gentype=None):
    """ Identifying a name in Elsewhere_Stmt node"""
    get_name_or_defer(stmt, node.items[1], res_value, config)

def search_Suffix(stmt, node, config, gentype=None):
    """ Identifying a name in Suffix node"""
    get_name_or_defer(stmt, node.items[0], res_value, config)
    get_name_or_defer(stmt, node.items[1], res_value, config)

def search_Declaration_Type_Spec(stmt, node, config, gentype=None):
    """ Identifying a name in Declaration_Type_Spec node"""
    get_name_or_defer(stmt, node.items[1], res_derivedtype, config)

def search_Data_Ref(stmt, node, config, gentype=None):
    """ Identifying a name in Data_Ref node"""
    from fortlab.kgutils import KGName

    # NOTE: to limit the scope of data saving in derived type,
    #       the last part_ref would be the one that has config, gentype=gentype
    if isinstance(node.items[0], Fortran2003.Name):
        get_name_or_defer(stmt, node.items[0], res_value, config, gentype=gentype)
    elif isinstance(node.items[0], Fortran2003.Part_Ref):
        get_name_or_defer(stmt, node.items[0].items[0], res_value, config, gentype=gentype) 
        get_name_or_defer(stmt, node.items[0].items[1], res_value, config) 

    for item in node.items[1:]:
        if isinstance(item, Fortran2003.Name): pass
        elif isinstance(item, Fortran2003.Part_Ref):
            get_name_or_defer(stmt, item.items[1], res_value, config)
        elif item is None: pass
        else: raise ProgramException('Unknown type: %s'%item.__class)

def search_Structure_Constructor(stmt, node, config, gentype=None):
    """ Identifying a name in Structure_Constructor node"""
    #get_name_or_defer(stmt, node.items[0], res_derivedtype)
    # NOTE: parser found ordinary subprogram as Structure_Constructor
    get_name_or_defer(stmt, node.items[0], res_value + res_derivedtype, config)
    get_name_or_defer(stmt, node.items[1], res_value, config)

def search_Binary_Constant(stmt, node, config, gentype=None):
    """ Identifying a name in Binary_Constant node"""
    pass

def search_Octal_Constant(stmt, node, config, gentype=None):
    """ Identifying a name in Octal_Constant node"""
    pass

def search_Hex_Constant(stmt, node, config, gentype=None):
    """ Identifying a name in Hex_Constant node"""
    pass

def search_Intrinsic_Stmt(stmt, node, config, gentype=None):
    """ Identifying a name in Intrinsic_Stmt node"""
    pass
    #get_name_or_defer(stmt, node.items[1], res_subprogram)

def search_Derived_Type_Stmt(stmt, node, config, gentype=None):
    """ Identifying a name in Derived_Type_Stmt node"""
    get_name_or_defer(stmt, node.items[0], res_value, config)
    get_name_or_defer(stmt, node.items[2], res_value, config)

def search_Access_Stmt(stmt, node, config, gentype=None):
    """ Identifying a name in Access_Stmt node"""
    get_name_or_defer(stmt, node.items[1], res_anything, config)

def search_Function_Reference(stmt, node, config, gentype=None):
    """ Identifying a name in Function_Reference node"""
    get_name_or_defer(stmt, node.items[0], res_function, config)
    get_name_or_defer(stmt, node.items[1], res_value, config)

def search_Return_Stmt(stmt, node, config, gentype=None):
    """ Identifying a name in Return_Stmt node"""
    get_name_or_defer(stmt, node.items[0], res_function, config)

def search_Print_Stmt(stmt, node, config, gentype=None):
    """ Identifying a name in Print_Stmt node"""
    get_name_or_defer(stmt, node.items[0], res_value, config)
    get_name_or_defer(stmt, node.items[1], res_value, config)

def search_Format(stmt, node, config, gentype=None):
    """ Identifying a name in Format node"""
    if hasattr(node, 'items') and len(node.items)>0:
        get_name_or_defer(stmt, node.items[0], res_value, config)

def search_Implicit_Stmt(stmt, node, config, gentype=None):
    """ Identifying a name in Implicit_Stmt node"""
    if hasattr(node, 'items') and len(node.items)>0:
        get_name_or_defer(stmt, node.items[0], res_value, config)

def search_Exit_Stmt(stmt, node, config, gentype=None):
    """ Identifying a name in Exit_Stmt node"""
    pass

def search_Pointer_Assignment_Stmt(stmt, node, config, gentype=None):
    """ Identifying a name in Pointer_Assignment_Stmt node"""
    get_name_or_defer(stmt, node.items[0], res_ptr_object, config) # data pointer obj or procedure pointer obj
    get_name_or_defer(stmt, node.items[2], res_target, config) # data target or procedure target

def search_Proc_Component_Ref(stmt, node, config, gentype=None):
    """ Identifying a name in Proc_Component_Ref node"""
    get_name_or_defer(stmt, node.items[0], res_value, config, gentype=gentype)
    # Type definition may handle a procedure component name?
    #get_name_or_defer(stmt, node.items[2], res_value)

def search_Io_Unit(stmt, node, config, gentype=None):
    """ Identifying a name in Io_Unit node"""
    if hasattr(node, 'items') and len(node.items)>0:
        get_name_or_defer(stmt, node.items[0], res_value, config)

def search_Level_3_Expr(stmt, node, config, gentype=None):
    """ Identifying a name in Level_3_Expr node"""
    get_name_or_defer(stmt, node.items[0], res_value, config)
    get_name_or_defer(stmt, node.items[2], res_value, config)

def search_Open_Stmt(stmt, node, config, gentype=None):
    """ Identifying a name in Open_Stmt node"""
    get_name_or_defer(stmt, node.items[0], res_value, config)
    get_name_or_defer(stmt, node.items[1], res_value, config)

def search_Connect_Spec(stmt, node, config, gentype=None):
    """ Identifying a name in Connect_Spec node"""
    get_name_or_defer(stmt, node.items[0], res_value, config)
    get_name_or_defer(stmt, node.items[1], res_value, config)

def search_Endfile_Stmt(stmt, node, config, gentype=None):
    """ Identifying a name in Endfile_Stmt node"""
    get_name_or_defer(stmt, node.items[0], res_value, config)
    get_name_or_defer(stmt, node.items[1], res_value, config)

def search_Position_Spec(stmt, node, config, gentype=None):
    """ Identifying a name in Position_Spec node"""
    get_name_or_defer(stmt, node.items[0], res_value, config)
    get_name_or_defer(stmt, node.items[1], res_value, config)

def search_Close_Stmt(stmt, node, config, gentype=None):
    """ Identifying a name in Close_Stmt node"""
    get_name_or_defer(stmt, node.items[0], res_value, config)
    get_name_or_defer(stmt, node.items[1], res_value, config)

def search_Close_Spec(stmt, node, config, gentype=None):
    """ Identifying a name in Close_Spec node"""
    get_name_or_defer(stmt, node.items[0], res_value, config)
    get_name_or_defer(stmt, node.items[1], res_value, config)

def search_Label(stmt, node, config, gentype=None):
    """ Identifying a name in Label node"""
    pass

def search_Io_Implied_Do(stmt, node, config, gentype=None):
    """ Identifying a name in Io_Implied_Do node"""
    get_name_or_defer(stmt, node.items[0], res_value, config)
    get_name_or_defer(stmt, node.items[1], res_value, config)

def search_Io_Implied_Do_Control(stmt, node, config, gentype=None):
    """ Identifying a name in Io_Implied_Do_Control node"""
    get_name_or_defer(stmt, node.items[0], res_typedecl, config, gentype=KGGenType.STATE_OUT)
    get_name_or_defer(stmt, node.items[1], res_value, config)
    get_name_or_defer(stmt, node.items[2], res_value, config)
    get_name_or_defer(stmt, node.items[3], res_value, config)

def search_Format_Stmt(stmt, node, config, gentype=None):
    """ Identifying a name in Format_Stmt node"""
    get_name_or_defer(stmt, node.items[0], res_value, config)
    # No need for searching format-items?
    #get_name_or_defer(stmt, node.items[1], res_value)

def search_Format_Specification(stmt, node, config, gentype=None):
    """ Identifying a name in Format_Specification node"""
    get_name_or_defer(stmt, node.items[0], res_value, config)
    get_name_or_defer(stmt, node.items[1], res_value, config)
    get_name_or_defer(stmt, node.items[2], res_value, config)

def search_Format_Item_C1002(stmt, node, config, gentype=None):
    """ Identifying a name in Format_Item_C1002 node"""
    get_name_or_defer(stmt, node.items[0], res_value, config)
    get_name_or_defer(stmt, node.items[1], res_value, config)

def search_Control_Edit_Desc(stmt, node, config, gentype=None):
    """ Identifying a name in Control_Edit_Desc node"""
    get_name_or_defer(stmt, node.items[0], res_value, config)
    get_name_or_defer(stmt, node.items[1], res_value, config)

def search_Format_Item(stmt, node, config, gentype=None):
    """ Identifying a name in Format_Item node"""
    get_name_or_defer(stmt, node.items[0], res_value, config)
    get_name_or_defer(stmt, node.items[1], res_value, config)

def search_Alloc_Opt(stmt, node, config, gentype=None):
    """ Identifying a name in Alloc_Opt node"""
    get_name_or_defer(stmt, node.items[0], res_value, config)
    get_name_or_defer(stmt, node.items[1], res_value, config)

def search_Deallocate_Stmt(stmt, node, config, gentype=None):
    """ Identifying a name in Deallocate_Stmt node"""
    get_name_or_defer(stmt, node.items[0], res_value, config)
    get_name_or_defer(stmt, node.items[1], res_value, config)

def search_Cycle_Stmt(stmt, node, config, gentype=None):
    """ Identifying a name in Cycle_Stmt node"""
    pass

def search_External_Stmt(stmt, node, config, gentype=None):
    """ Identifying a name in External_Stmt node"""
    get_name_or_defer(stmt, node.items[1], \
        [ TypeDeclarationStatement, Function, Subroutine ], config)

def search_Case_Value_Range(stmt, node, config, gentype=None):
    """ Identifying a name in Case_Value_Range node"""
    get_name_or_defer(stmt, node.items[0], res_value, config)
    get_name_or_defer(stmt, node.items[1], res_value, config)

def search_Forall_Construct_Stmt(stmt, node, config, gentype=None):
    """ Identifying a name in Forall_Construct_Stmt node"""
    get_name_or_defer(stmt, node.items[1], res_value, config)

def search_Forall_Header(stmt, node, config, gentype=None):
    """ Identifying a name in Forall_Header node"""
    get_name_or_defer(stmt, node.items[0], res_value, config)
    get_name_or_defer(stmt, node.items[1], res_value, config)

def search_Forall_Triplet_Spec(stmt, node, config, gentype=None):
    """ Identifying a name in Forall_Triplet_Spec node"""
    get_name_or_defer(stmt, node.items[0], res_value, config)
    get_name_or_defer(stmt, node.items[1], res_value, config)
    get_name_or_defer(stmt, node.items[2], res_value, config)
    get_name_or_defer(stmt, node.items[3], res_value, config)

def search_Goto_Stmt(stmt, node, config, gentype=None):
    """ Identifying a name in Goto_Stmt node"""
    pass

def search_Continue_Stmt(stmt, node, config, gentype=None):
    """ Identifying a name in Continue_Stmt node"""
    pass

def search_Wait_Stmt(stmt, node, config, gentype=None):
    """ Identifying a name in Wait_Stmt node"""
    get_name_or_defer(stmt, node.items[1], res_value, config)

def search_Wait_Spec(stmt, node, config, gentype=None):
    """ Identifying a name in Wait_Spec node"""
    if hasattr(node, 'items') and len(node.items)>0:
        for item in node.items:
            get_name_or_defer(stmt, item, res_value, config)

def search_Rewind_Stmt(stmt, node, config, gentype=None):
    """ Identifying a name in Rewind_Stmt node"""
    get_name_or_defer(stmt, node.items[0], res_value, config)
    get_name_or_defer(stmt, node.items[1], res_value, config)

def search_Flush_Stmt(stmt, node, config, gentype=None):
    """ Identifying a name in Flush_Stmt node"""
    get_name_or_defer(stmt, node.items[0], res_value, config)
    get_name_or_defer(stmt, node.items[1], res_value, config)

def search_Import_Stmt(stmt, node, config, gentype=None):
    """ Identifying a name in Import_Stmt node"""
    get_name_or_defer(stmt, node.items[1], res_anything, config)

def search_Block_Data_Stmt(stmt, node, config, gentype=None):
    """ Identifying a name in Block_Data_Stmt node"""
    # NOTE: Temporary solution
    pass

def search_Data_Stmt(stmt, node, config, gentype=None):
    """ Identifying a name in Data_Stmt node"""
    if hasattr(node, 'items') and len(node.items)>0:
        for item in node.items:
            get_name_or_defer(stmt, item, res_typedecl, config)

def search_Data_Stmt_Value(stmt, node, config, gentype=None):
    """ Identifying a name in Data_Stmt_Value node"""
    get_name_or_defer(stmt, node.items[0], res_typedecl, config)
    get_name_or_defer(stmt, node.items[1], res_typedecl, config)

def search_Save_Stmt(stmt, node, config, gentype=None):
    """ Identifying a name in Save_Stmt node"""
    get_name_or_defer(stmt, node.items[1], res_typedecl, config)

def search_Asynchronous_Stmt(stmt, node, config, gentype=None):
    """ Identifying a name in Asynchronous_Stmt node"""
    get_name_or_defer(stmt, node.items[1], res_typedecl, config)

def search_Allocatable_Stmt(stmt, node, config, gentype=None):
    """ Identifying a name in Allocatable_Stmt node"""
    get_name_or_defer(stmt, node.items[1], res_typedecl, config)

def search_Common_Stmt(stmt, node, config, gentype=None):
    """ Identifying a name in Common_Stmt node"""
    if hasattr(node, 'items') and len(node.items)>0:
        for itemlist in node.items:
            for name, _item in itemlist:
                get_name_or_defer(stmt, _item, res_value, config)

def search_Data_Stmt_Set(stmt, node, config, gentype=None):
    """ Identifying a name in Data_Stmt_Set node"""
    get_name_or_defer(stmt, node.items[0], res_value, config)
    get_name_or_defer(stmt, node.items[1], res_value, config)

def search_Dimension_Stmt(stmt, node, config, gentype=None):
    """ Identifying a name in Dimension_Stmt node"""
    if hasattr(node, 'items') and len(node.items)>0:
        for itemlist in node.items:
            for name, _item in itemlist:
                get_name_or_defer(stmt, _item, res_value, config)

def search_Equivalence_Stmt(stmt, node, config, gentype=None):
    """ Identifying a name in Equivalence_Stmt node"""
    get_name_or_defer(stmt, node.items[1], res_typedecl, config)

def search_Equivalence_Set(stmt, node, config, gentype=None):
    """ Identifying a name in Equivalence_Set node"""
    get_name_or_defer(stmt, node.items[0], res_typedecl, config)
    get_name_or_defer(stmt, node.items[1], res_typedecl, config)

def search_Intent_Stmt(stmt, node, config, gentype=None):
    """ Identifying a name in Intent_Stmt node"""
    #get_name_or_defer(stmt, node.items[0], res_typedecl)
    get_name_or_defer(stmt, node.items[1], res_typedecl, config)

def search_Intent_Spec(stmt, node, config, gentype=None):
    """ Identifying a name in Intent_Spec node"""
    pass

def search_Namelist_Stmt(stmt, node, config, gentype=None):
    """ Identifying a name in Namelist_Stmt node"""
    if hasattr(node, 'items') and len(node.items)>0:
        for nlname, nlgroup in node.items:
            get_name_or_defer(stmt, nlgroup, res_typedecl, config)

def search_Optional_Stmt(stmt, node, config, gentype=None):
    """ Identifying a name in Optional_Stmt node"""
    get_name_or_defer(stmt, node.items[1], res_typedecl, config)

def search_Pointer_Stmt(stmt, node, config, gentype=None):
    """ Identifying a name in Pointer_Stmt node"""
    get_name_or_defer(stmt, node.items[1], res_anything, config)

def search_Protected_Stmt(stmt, node, config, gentype=None):
    """ Identifying a name in Protected_Stmt node"""
    get_name_or_defer(stmt, node.items[1], res_typedecl, config)

def search_Target_Stmt(stmt, node, config, gentype=None):
    """ Identifying a name in Target_Stmt node"""
    get_name_or_defer(stmt, node.items[0], res_anything, config)

def search_Target_Entity_Decl(stmt, node, config, gentype=None):
    """ Identifying a name in Target_Entity_Decl node"""
    get_name_or_defer(stmt, node.items[0], res_anything, config)
    defer(stmt, node.items[1], config)
    get_name_or_defer(stmt, node.items[2], res_value, config)
    get_name_or_defer(stmt, node.items[3], res_value, config) 

def search_Volatile_Stmt(stmt, node, config, gentype=None):
    """ Identifying a name in Volatile_Stmt node"""
    get_name_or_defer(stmt, node.items[1], res_anything, config)

def search_Value_Stmt(stmt, node, config, gentype=None):
    """ Identifying a name in Value_Stmt node"""
    get_name_or_defer(stmt, node.items[1], res_typedecl, config)

def search_Backspace_Stmt(stmt, node, config, gentype=None):
    """ Identifying a name in Backspace_Stmt node"""
    get_name_or_defer(stmt, node.items[0], res_value, config)
    get_name_or_defer(stmt, node.items[1], res_value, config)

def search_Forall_Stmt(stmt, node, config, gentype=None):
    """ Identifying a name in Forall_Stmt node"""
    get_name_or_defer(stmt, node.items[0], res_value, config)
    get_name_or_defer(stmt, node.items[1], res_value, config)

def search_Inquire_Stmt(stmt, node, config, gentype=None):
    """ Identifying a name in Inquire_Stmt node"""
    get_name_or_defer(stmt, node.items[0], res_value, config)
    get_name_or_defer(stmt, node.items[1], res_value, config)
    get_name_or_defer(stmt, node.items[2], res_value, config)

def search_Inquire_Spec(stmt, node, config, gentype=None):
    """ Identifying a name in Inquire_Spec node"""
    get_name_or_defer(stmt, node.items[0], res_value, config)
    get_name_or_defer(stmt, node.items[1], res_value, config)

def search_Nullify_Stmt(stmt, node, config, gentype=None):
    """ Identifying a name in Nullify_Stmt node"""
    get_name_or_defer(stmt, node.items[0], res_value, config)
    get_name_or_defer(stmt, node.items[1], res_value, config)

def search_Where_Stmt(stmt, node, config, gentype=None):
    """ Identifying a name in Where_Stmt node"""
    get_name_or_defer(stmt, node.items[0], res_value, config)
    get_name_or_defer(stmt, node.items[1], res_value, config)

def search_Arithmetic_If_Stmt(stmt, node, config, gentype=None):
    """ Identifying a name in Arithmetic_If_Stmt node"""
    get_name_or_defer(stmt, node.items[0], res_value, config)
    get_name_or_defer(stmt, node.items[1], res_value, config)
    get_name_or_defer(stmt, node.items[2], res_value, config)
    get_name_or_defer(stmt, node.items[3], res_value, config)

def search_Computed_Goto_Stmt(stmt, node, config, gentype=None):
    """ Identifying a name in Computed_Goto_Stmt node"""
    get_name_or_defer(stmt, node.items[0], res_value, config)
    get_name_or_defer(stmt, node.items[1], res_value, config)

def search_Actual_Arg_Spec(stmt, node, config, gentype=None):
    """ Identifying a name in Actual_Arg_Spec node"""
    get_name_or_defer(stmt, node.items[1], res_value, config)

def search_Data_Pointer_Object(stmt, node, config, gentype=None):
    """ Identifying a name in Data_Pointer_Object node"""

    get_name_or_defer(stmt, node.items[0], res_value, config, gentype=gentype)

    if node.items[2] and not isinstance(node.items[2], Fortran2003.Name):
        get_name_or_defer(stmt, node.items[2], res_value, config)

def search_Type_Attr_Spec(stmt, node, config, gentype=None):
    """ Identifying a name in Type_Attr_Spec node"""
    if isinstance(node.items[0], str) and node.items[0]=='EXTENDS':
        get_name_or_defer(stmt, node.items[1], res_derivedtype, config)
    else:
        get_name_or_defer(stmt, node.items[0], res_value, config)
        get_name_or_defer(stmt, node.items[1], res_value, config)

def search_Level_5_Expr(stmt, node, config, gentype=None):
    """ Identifying a name in Level_5_Expr node"""
    get_name_or_defer(stmt, node.items[0], res_value, config)
    get_name_or_defer(stmt, node.items[2], res_value, config)

def search_Parameter_Stmt(stmt, node, config, gentype=None):
    """ Identifying a name in Parameter_Stmt node"""
    get_name_or_defer(stmt, node.items[1], res_value, config)
   
def search_Named_Constant_Def(stmt, node, config, gentype=None): 
    """ Identifying a name in Named_Constant_Def node"""
    get_name_or_defer(stmt, node.items[0], res_value, config)
    get_name_or_defer(stmt, node.items[1], res_value, config)

def search_Saved_Entity(stmt, node, config, gentype=None):
    """ Identifying a name in Saved_Entity node"""
    if len(node.items)==3 and node.items[0]=='/' and node.items[2]=='/':
        get_name_or_defer(stmt, node.items[1], res_common, config)
    else:
        for item in node.items:
            get_name_or_defer(stmt, item, res_value, config)

def search_Alloc_Opt(stmt, node, config, gentype=None):
    """ Identifying a name in Alloc_Opt node"""
    get_name_or_defer(stmt, node.items[1], res_value, config)

def search_Dealloc_Opt(stmt, node, config, gentype=None):
    """ Identifying a name in Dealloc_Opt node"""
    get_name_or_defer(stmt, node.items[1], res_value, config)

def search_Generic_Spec(stmt, node, config, gentype=None):
    """ Identifying a name in Generic_Spec node"""
    pass

def search_Assumed_Size_Spec(stmt, node, config, gentype=None):
    """ Identifying a name in Assumed_Size_Spec node"""
    get_name_or_defer(stmt, node.items[0], res_value, config)
    get_name_or_defer(stmt, node.items[1], res_value, config)

def search_Common_Block_Object(stmt, node, config, gentype=None):
    """ Identifying a name in Common_Block_Object node"""
    get_name_or_defer(stmt, node.items[0], res_value, config)
    get_name_or_defer(stmt, node.items[1], res_value, config)

def search_Ac_Implied_Do(stmt, node, config, gentype=None):
    """ Identifying a name in Ac_Implied_Do node"""
    get_name_or_defer(stmt, node.items[0], res_value, config)
    get_name_or_defer(stmt, node.items[1], res_value, config)

def search_Ac_Implied_Do_Control(stmt, node, config, gentype=None):
    """ Identifying a name in Ac_Implied_Do_Control node"""
    get_name_or_defer(stmt, node.items[0], res_value, config, gentype=KGGenType.STATE_OUT)
    if node.items[1]:
        for item in node.items[1]:
            get_name_or_defer(stmt, item, res_value, config)

def search_Specific_Binding(stmt, node, config, gentype=None):
    """ Identifying a name in Specific_Binding node"""
    get_name_or_defer(stmt, node.items[0], res_typespec + [ Interface ], config)
    get_name_or_defer(stmt, node.items[1], res_value, config)
    if not hasattr(node.items[1], 'string') or 'DEFERRED' not in node.items[1].string:
        if node.items[3] is None:
            get_name_or_defer(stmt, node.items[2], res_subprogram, config)
        else:
            get_name_or_defer(stmt, node.items[3], res_subprogram, config)

def search_Binding_Attr(stmt, node, config, gentype=None):
    """ Identifying a name in Binding_Attr node"""
    pass

def search_Masked_Elsewhere_Stmt(stmt, node, config, gentype=None):
    """ Identifying a name in Masked_Elsewhere_Stmt node"""
    get_name_or_defer(stmt, node.items[0], res_value, config)

def search_Procedure_Designator(stmt, node, config, gentype=None):
    """ Identifying a name in Procedure_Designator node"""
    get_name_or_defer(stmt, node.items[0], res_value, config)

def search_Associate_Stmt(stmt, node, config, gentype=None):
    """ Identifying a name in Associate_Stmt node"""
    get_name_or_defer(stmt, node.items[1], res_value, config)

def search_Association(stmt, node, config, gentype=None):
    """ Identifying a name in Association node"""
    get_name_or_defer(stmt, node.items[2], res_value, config)

def search_Generic_Binding(stmt, node, config, gentype=None):
    """ Identifying a name in Generic_Binding node"""
    get_name_or_defer(stmt, node.items[0], res_value, config)
    get_name_or_defer(stmt, node.items[2], [ SpecificBinding ], config)

def search_Complex_Literal_Constant(stmt, node, config, gentype=None):
    """ Identifying a name in Complex_Literal_Constant node"""
    get_name_or_defer(stmt, node.items[0], res_value, config)
    get_name_or_defer(stmt, node.items[1], res_value, config)

def search_Char_Length(stmt, node, config, gentype=None):
    """ Identifying a name in Char_Length node"""
    get_name_or_defer(stmt, node.items[1], res_value, config)

def search_Data_Implied_Do(stmt, node, config, gentype=None):
    """ Identifying a name in Data_Implied_Do node"""
    get_name_or_defer(stmt, node.items[0], res_value, config)
    get_name_or_defer(stmt, node.items[1], res_value, config, gentype=KGGenType.STATE_OUT)
    get_name_or_defer(stmt, node.items[2], res_value, config)
    get_name_or_defer(stmt, node.items[3], res_value, config)
    get_name_or_defer(stmt, node.items[4], res_value, config)

def search_Ac_Spec(stmt, node, config, gentype=None):
    """ Identifying a name in Ac_Spec node"""
    defer(stmt, node.items[0], config)
    get_name_or_defer(stmt, node.items[1], res_value, config)

def search_Sequence_Stmt(stmt, node, config, gentype=None):
    """ Identifying a name in Sequence_Stmt node"""
    pass

def search_Stmt_Function_Stmt(stmt, node, config, gentype=None):
    """ Identifying a name in Stmt_Function_Stmt node"""
    get_name_or_defer(stmt, node.items[0], res_typedecl, config)

def search_Language_Binding_Spec(stmt, node, config, gentype=None):
    """ Identifying a name in Language_Binding_Spec node"""
    # No need to resolve exteranl c library routines
    pass

def search_Select_Type_Stmt(stmt, node, config, gentype=None):
    """ Identifying a name in search_Select_Type_Stmt node"""
    get_name_or_defer(stmt, node.items[1], res_typedecl, config)

def search_Type_Guard_Stmt(stmt, node, config, gentype=None):
    """ Identifying a name in search_Type_Guard_Stmt node"""
    get_name_or_defer(stmt, node.items[1], res_typespec, config)
    #show_tree(node)
    #import pdb ;pdb.set_trace()

def search_Implicit_Spec(stmt, node, config, gentype=None):
    """
    <implicit-spec> = <declaration-type-spec> ( <letter-spec-list> )
    """
    get_name_or_defer(stmt, node.items[0], res_typespec, config)
    get_name_or_defer(stmt, node.items[1], res_typespec, config)

def search_Letter_Spec(stmt, node, config, gentype=None):
    pass
    
def search_Procedure_Declaration_Stmt(stmt, node, config, gentype=None):

    get_name_or_defer(stmt, node.items[0], [Interface], config)
    get_name_or_defer(stmt, node.items[1], res_value, config)

    #show_tree(node)
    #import pdb ;pdb.set_trace()

def search_Binding_PASS_Arg_Name(stmt, node, config, gentype=None):

    #show_tree(node)
    #import pdb ;pdb.set_trace()

    pass
  
def search_Char_Selector(stmt, node, config, gentype=None):

    get_name_or_defer(stmt, node.items[0], res_value, config)
    get_name_or_defer(stmt, node.items[1], res_value, config)

def search_Enum_Def_Stmt(stmt, node, config, gentype=None):
    pass
  
def search_Enumerator_Def_Stmt(stmt, node, config, gentype=None):

    get_name_or_defer(stmt, node.items[1], res_value, config)
  
def search_Enumerator(stmt, node, config, gentype=None):

    get_name_or_defer(stmt, node.items[2], res_value, config)



"""Public API for Fortran parser.

Module content
--------------
"""
#Author: Pearu Peterson <pearu@cens.ioc.ee>
#Created: Oct 2006

__autodoc__ = ['get_reader', 'parse', 'walk']

import locale

#import Fortran2003
# import all Statement classes:
#from fortlab.resolver.base_classes import classes
from fortlab.resolver.base_classes import BeginStatement, EndStatement
from fortlab.resolver.block_statements import *

# CHAR_BIT is used to convert object bit sizes to byte sizes
from fortlab.resolver.utils import CHAR_BIT

def get_reader(input, isfree=None, isstrict=None, include_dirs = None, source_only = None,
               ignore_comments = True):
    """ Returns Fortran reader instance.

    Parameters
    ----------
    input : str
      Specify a string or filename containing Fortran code.    
    isfree, isstrict : {None, bool}
      Specify input Fortran format. The values are determined from the
      input. If that fails then isfree=True and isstrict=False is assumed.
    include_dirs : {None, list}
      Specify a list of include directories. The default list (when
      include_dirs=None) contains the current working directory and
      the directory of ``filename``.
    source_only : {None, list}
      Specify a list of Fortran file names that are searched when the
      ``USE`` statement is encountered.

    Returns
    -------
    reader : `FortranReader`

    Notes
    -----
    If ``input`` is a C filename then the functions searches for comment
    lines starting with ``/*f2py`` and reads following lines as PYF file
    content until a line ``*/`` is found.

    See also
    --------
    parse
    """
    import os
    import re
    from fortlab.resolver.readfortran import FortranFileReader, FortranStringReader
    if os.path.isfile(input):
        name,ext = os.path.splitext(input)
        if ext.lower() in ['.c']:
            # get signatures from C file comments starting with `/*f2py` and ending with `*/`.
            # TODO: improve parser to take line number offset making line numbers in
            #       parser messages correct.
            f2py_c_comments = re.compile('/[*]\s*f2py\s.*[*]/',re.I | re.M)
            f = open(filename,'r')
            c_input = ''
            for s1 in f2py_c_comments.findall(f.read()):
                c_input += s1[2:-2].lstrip()[4:] + '\n'
            f.close()
            if isfree is None: isfree = True
            if isstrict is None: isstrict = True
            return parse(c_input, isfree, isstrict, include_dirs)
        reader = FortranFileReader(input, include_dirs = include_dirs, source_only = source_only)
    elif isinstance(input, str):
        reader = FortranStringReader(input, include_dirs = include_dirs, source_only = source_only)
    elif type(input) == type(u""): # python 2
        enc = locale.getpreferredencoding(False)
        reader = FortranStringReader(input.encode(enc), include_dirs = include_dirs, source_only = source_only)
    else:
        raise TypeError('Expected string or filename input but got %s' % (type(input)))
    if isfree is None: isfree = reader.isfree
    if isstrict is None: isstrict = reader.isstrict
    reader.set_mode(isfree, isstrict)
    return reader

def parse(input, isfree=None, isstrict=None, include_dirs = None, source_only = None,
          ignore_comments = True, analyze=True):
    """ Parse input and return Statement tree.

    Parameters
    ----------
    input : str
      Specify a string or filename containing Fortran code.
    isfree, isstrict : {None, bool}
      Specify input Fortran format. The values are determined from the
      input. If that fails then isfree=True and isstrict=False is assumed.
    include_dirs : {None, list}
      Specify a list of include directories. The default list (when
      include_dirs=None) contains the current working directory and
      the directory of ``filename``.
    source_only : {None, list}
      Specify a list of Fortran file names that are searched when the
      ``USE`` statement is encountered.
    ignore_comments : bool
      When True then discard all comment lines in the Fortran code.
    analyze : bool
      When True then apply run analyze method on the Fortran code tree.

    Returns
    -------
    block : `fparser.api.BeginSource`

    Examples
    --------

        >>> code = '''
        ... c comment
        ...       subroutine foo(a)
        ...         integer a
        ...         print*, "a=",a
        ...       end
        ... '''
        >>> tree = parse(code,isfree=False)
        >>> print(tree)
              !BEGINSOURCE <cStringIO.StringI object at 0x1798030> mode=fix90
                SUBROUTINE foo(a)
                  INTEGER a
                  PRINT *, "a=", a
                END SUBROUTINE foo
        >>> print(str(tree))
              BeginSource
                blocktype='beginsource'
                name='<cStringIO.StringI object at 0x1798030> mode=fix90'
                a=AttributeHolder:
              external_subprogram=<dict with keys ['foo']>
                content:
                  Subroutine
                    args=['a']
                    item=Line('subroutine foo(a)',(3, 3),'')
                    a=AttributeHolder:
                variables=<dict with keys ['a']>
                    content:
                      Integer
                        selector=('', '')
                        entity_decls=['a']
                        item=Line('integer a',(4, 4),'')
                      Print
                        item=Line('print*, "a=",a',(5, 5),'')
                  EndSubroutine
                    blocktype='subroutine'
                    name='foo'
            item=Line('end',(6, 6),'')

    See also
    --------
    get_reader
    """
    from fortlab.resolver.parsefortran import FortranParser
    reader = get_reader(input, isfree, isstrict, include_dirs, source_only)
    parser = FortranParser(reader, ignore_comments = ignore_comments)
    parser.parse()
    if analyze:
        parser.analyze()
    return parser.block

def walk(stmt, depth=-1, _initial_depth = None):
    """ Generate Fortran statements by walking the stmt tree until given depth.

    For each block statement in stmt, the walk functions yields a
    tuple ``(statement, depth)`` where ``depth`` is the depth of tree
    stucture for statement.

    Parameters
    ----------
    stmt : Statement
    depth : int
      If depth is positive then walk in the tree until given depth.
      If depth is negative then walk the whole tree.

    Returns
    -------
    generator

    Examples
    --------

      ::

        from fparser import api
        source_str = '''
        subroutine foo
          integer i, r
          do i=1,100
            r = r + i
          end do
        end
        '''
        tree = api.parse(source_str)
        for stmt, depth in api.walk(tree):
            print(depth, stmt.item)

      that will print::

        1 line #2'subroutine foo'
        2 line #3'integer i, r'
        2 line #4'do i=1,100'
        3 line #5'r = r + i'
        2 line #6'end do'
        1 line #7'end'

    """
    if _initial_depth is None:
        if depth==0:
            return
        _initial_depth = depth
    if not isinstance(stmt, BeginSource):
        yield stmt, _initial_depth - depth
    if isinstance(stmt, BeginStatement):
        last_stmt = stmt.content[-1]
        last_index = len(stmt.content)
        if isinstance(last_stmt, EndStatement):
            last_index -= 1
        else:
            last_stmt = None
        if depth != 0:
            for substmt in stmt.content[:last_index]:
                if stmt.__class__.__name__ in [ 'If' ]: continue # KGEN addition
                for statement, statement_depth in walk(substmt, depth-1, _initial_depth):
                    yield statement, statement_depth
        if last_stmt is not None:
            yield last_stmt, _initial_depth - depth



# gen_write_type.py
from __future__ import absolute_import
 
from fortlab.resolver import statements, block_statements, typedecl_statements
from fortlab.kgplugin import Kgen_Plugin
from .verify_utils import get_dtype_verifyname, get_typedecl_verifyname, kernel_verify_contains, kernel_verify_kgenutils, is_remove_state, \
    is_zero_array

class Verify_Type(Kgen_Plugin):
    def __init__(self):
        self.frame_msg = None

        self.kernel_created_use_items = []
        self.kernel_created_public_items = []

    # registration
    def register(self, msg):
        self.frame_msg = msg

        # register initial events
        self.frame_msg.add_event(KERNEL_SELECTION.ALL, FILE_TYPE.KERNEL, GENERATION_STAGE.BEGIN_PROCESS, \
            block_statements.Type, self.has_state_out_info, self.create_dtype_verify_subr) 

        self.frame_msg.add_event(KERNEL_SELECTION.ALL, FILE_TYPE.KERNEL, GENERATION_STAGE.BEGIN_PROCESS, \
            statements.Use, self.has_dtype_res_path, self.add_verifynames_in_use_public) 

    def has_state_out_info(self, node):
        if node.kgen_stmt and 'abstract' not in node.kgen_stmt.specs and \
            hasattr(node.kgen_stmt, 'geninfo') and len(node.kgen_stmt.geninfo)>0:
            return True
        else: return False

    def has_dtype_res_path(self, node):
        if node.kgen_stmt and hasattr(node.kgen_stmt, 'geninfo'):
            if not node.kgen_stmt.isonly: return False
            for gentype, reqlist in node.kgen_stmt.geninfo.items():
                if any(len(req.res_stmts)>0 and isinstance(req.res_stmts[0], block_statements.Type) and \
                    'abstract' not in req.res_stmts[0].specs for uname, req in reqlist):
                    return True
        return False

    def add_verifynames_in_use_public(self, node):
        parent = node.kgen_parent

        for gentype, reqlist in node.kgen_stmt.geninfo.items():
            for uname, req in reqlist:
                if len(req.res_stmts)>0 and isinstance(req.res_stmts[0], block_statements.Type):
                    subrname = get_dtype_verifyname(req.res_stmts[0])
                    checks = lambda n: n.kgen_match_class==statements.Use and n.kgen_stmt and n.kgen_stmt.isonly and \
                        subrname in n.kgen_stmt.items
                    if (id(parent),subrname) not in self.kernel_created_use_items and not part_has_node(parent, USE_PART, checks):
                        attrs = {'name':node.kgen_stmt.name, 'isonly': True, 'items':[subrname]}
                        part_append_genknode(parent, USE_PART, statements.Use, attrs=attrs)
                        self.kernel_created_use_items.append((id(parent),subrname))

                    checks = lambda n: isinstance(n.kgen_stmt, statements.Public) and n.kgen_stmt.items and subrname in n.kgen_stmt.items
                    if (id(parent),subrname) not in self.kernel_created_public_items and isinstance(parent.kgen_stmt, block_statements.Module) and \
                        not part_has_node(parent, DECL_PART, checks):
                        attrs = {'items':[subrname]}
                        part_append_genknode(parent, DECL_PART, statements.Public, attrs=attrs)
                        self.kernel_created_public_items.append((id(parent),subrname))

    # process function
    def create_dtype_verify_subr(self, node):
        assert node.kgen_stmt, 'None kgen statement'

        def print_numarr_detail(parent, entity_name):
            attrs = {'items': ['count( var%%%s /= kgenref_var%%%s)'%(entity_name, entity_name), \
                '" of "', 'size( var%%%s )'%entity_name, '" elements are different."']}
            part_append_genknode(parent, EXEC_PART, statements.Write, attrs=attrs)

            attrs = {'items': ['"Average - kernel "', 'sum(var%%%s)/real(size(var%%%s))'%(entity_name, entity_name)]}
            part_append_genknode(parent, EXEC_PART, statements.Write, attrs=attrs)

            attrs = {'items': ['"Average - reference "', 'sum(kgenref_var%%%s)/real(size(kgenref_var%%%s))'%(entity_name, entity_name)]}
            part_append_genknode(parent, EXEC_PART, statements.Write, attrs=attrs)

            attrs = {'items': ['"RMS of difference is "', 'rmsdiff_%s'%entity_name]}
            part_append_genknode(parent, EXEC_PART, statements.Write, attrs=attrs)

            attrs = {'items': ['"Normalized RMS of difference is "', 'nrmsdiff_%s'%entity_name]}
            part_append_genknode(parent, EXEC_PART, statements.Write, attrs=attrs)

            attrs = {'items': ['""']}
            part_append_genknode(parent, EXEC_PART, statements.Write, attrs=attrs)

        def print_dtypearr_detail(parent, entity_name):

            attrs = {'items': [ '"    number of elements         : "', 'comp_check_status%numtotal']}
            part_append_genknode(parent, EXEC_PART, statements.Write, attrs=attrs)

            attrs = {'items': [ '"    identical                  : "', 'comp_check_status%numidentical']}
            part_append_genknode(parent, EXEC_PART, statements.Write, attrs=attrs)

            attrs = {'items': [ '"    not identical - out of tol.: "', 'comp_check_status%numouttol']}
            part_append_genknode(parent, EXEC_PART, statements.Write, attrs=attrs)

            attrs = {'items': [ '"    not identical - within tol.: "', 'comp_check_status%numintol']}
            part_append_genknode(parent, EXEC_PART, statements.Write, attrs=attrs)

            attrs = {'items': ['""']}
            part_append_genknode(parent, EXEC_PART, statements.Write, attrs=attrs)

        def print_dtype_detail(parent, entity_name):

            attrs = {'items': [ '"    number of components       : "', 'comp_check_status%numtotal']}
            part_append_genknode(parent, EXEC_PART, statements.Write, attrs=attrs)

            attrs = {'items': [ '"    identical                  : "', 'comp_check_status%numidentical']}
            part_append_genknode(parent, EXEC_PART, statements.Write, attrs=attrs)

            attrs = {'items': [ '"    not identical - out of tol.: "', 'comp_check_status%numouttol']}
            part_append_genknode(parent, EXEC_PART, statements.Write, attrs=attrs)

            attrs = {'items': [ '"    not identical - within tol.: "', 'comp_check_status%numintol']}
            part_append_genknode(parent, EXEC_PART, statements.Write, attrs=attrs)

            attrs = {'items': ['""']}
            part_append_genknode(parent, EXEC_PART, statements.Write, attrs=attrs)

        def print_num_detail(parent, entity_name):
            attrs = {'items': ['"Difference is "', 'diff_%s'%entity_name]}
            part_append_genknode(parent, EXEC_PART, statements.Write, attrs=attrs)

            attrs = {'items': ['""']}
            part_append_genknode(parent, EXEC_PART, statements.Write, attrs=attrs)

        def print_dummy_detail(parent, entity_name):
            attrs = {'items': ['"NOT IMPLEMENTED YET"']}
            part_append_genknode(parent, EXEC_PART, statements.Write, attrs=attrs)

            attrs = {'items': ['""']}
            part_append_genknode(parent, EXEC_PART, statements.Write, attrs=attrs)

        subrname = get_dtype_verifyname(node.kgen_stmt)
        if subrname is None: return

        parent = node.kgen_parent
        checks = lambda n: isinstance(n.kgen_stmt, block_statements.Subroutine) and n.name==subrname
        if  not part_has_node(parent, SUBP_PART, checks):

            checks = lambda n: n.kgen_isvalid and n.kgen_match_class==statements.Contains
            if not parent in kernel_verify_contains and not part_has_node(parent, CONTAINS_PART, checks):
                part_append_comment(parent, CONTAINS_PART, '')
                part_append_genknode(parent, CONTAINS_PART, statements.Contains)
                part_append_comment(parent, CONTAINS_PART, '')
                kernel_verify_contains.append(parent)

            checks = lambda n: n.kgen_isvalid and n.kgen_match_class==statements.Use and n.name=='kgen_utils_mod' and 'check_t' in n.items
            if not parent in kernel_verify_kgenutils and not part_has_node(parent, USE_PART, checks):

                attrs = {'name': 'kgen_utils_mod', 'isonly': True, 'items': ['check_t', 'kgen_init_check', 'kgen_init_verify',
                    'kgen_tolerance', 'kgen_minvalue', 'kgen_verboselevel', 'CHECK_IDENTICAL', 'CHECK_IN_TOL', 'CHECK_OUT_TOL']}
                part_append_genknode(parent, USE_PART, statements.Use, attrs=attrs)
                kernel_verify_kgenutils.append(parent)

            part_append_comment(parent, SUBP_PART, 'verify state subroutine for %s'%subrname)
            attrs = {'prefix': 'RECURSIVE', 'name': subrname, 'args': ['varname', 'check_status', 'var', 'kgenref_var']}
            subrobj = part_append_genknode(parent, SUBP_PART, block_statements.Subroutine, attrs=attrs)
            part_append_comment(parent, SUBP_PART, '')

            # varname
            attrs = {'type_spec': 'CHARACTER', 'attrspec': ['INTENT(IN)'], 'selector':('*', None), 'entity_decls': ['varname']}
            part_append_genknode(subrobj, DECL_PART, typedecl_statements.Character, attrs=attrs)

            # check_status
            attrs = {'type_spec': 'TYPE', 'attrspec': ['INTENT(INOUT)'], 'selector':(None, 'check_t'), 'entity_decls': ['check_status']}
            part_append_genknode(subrobj, DECL_PART, typedecl_statements.Type, attrs=attrs)

            # variable
            attrs = {'type_spec': 'TYPE', 'attrspec': ['INTENT(IN)'], 'selector':(None, node.name), 'entity_decls': ['var', 'kgenref_var']}
            part_append_genknode(subrobj, DECL_PART, typedecl_statements.Type, attrs=attrs)

            # dtype_check_status
            attrs = {'type_spec': 'TYPE', 'selector':(None, 'check_t'), 'entity_decls': ['dtype_check_status', 'comp_check_status']}
            part_append_genknode(subrobj, DECL_PART, typedecl_statements.Type, attrs=attrs)

            # check result
            attrs = {'type_spec': 'INTEGER', 'entity_decls': ['check_result']}
            part_append_genknode(subrobj, DECL_PART, typedecl_statements.Integer, attrs=attrs)

            # print check result
            attrs = {'type_spec': 'LOGICAL', 'entity_decls': ['is_print = .FALSE.']}
            part_append_genknode(subrobj, DECL_PART, typedecl_statements.Logical, attrs=attrs)
            part_append_comment(subrobj, DECL_PART, '')

            part_append_comment(subrobj, EXEC_PART, '')
            attrs = {'variable': 'check_status%numTotal', 'sign': '=', 'expr': 'check_status%numTotal + 1'}
            part_append_genknode(subrobj, EXEC_PART, statements.Assignment, attrs=attrs)
            part_append_comment(subrobj, EXEC_PART, '')

            attrs = {'designator': 'kgen_init_check', 'items': ['dtype_check_status', 'rank=check_status%rank']}
            part_append_genknode(subrobj, EXEC_PART, statements.Call, attrs=attrs)

            #comp_part = get_part(node, TYPE_COMP_PART) 
            comp_part = get_part(node, TYPE_PART) 
            for item in comp_part:
                if not hasattr(item, 'kgen_stmt'): continue
                if not isinstance(item.kgen_stmt, typedecl_statements.TypeDeclarationStatement): continue

                stmt = item.kgen_stmt
                entity_names = [ get_entity_name(decl) for decl in stmt.entity_decls ]
                topobj = subrobj
                for entity_name, entity_decl in zip(entity_names, stmt.entity_decls):
                    var = stmt.get_variable(entity_name)

                    if is_remove_state(entity_name, stmt): continue
                    if var.is_array() and is_zero_array(var, stmt): continue

                    callname = get_typedecl_verifyname(stmt, entity_name)

                    ifallocobj = None
                    ifassocobj = None
                    if var.is_allocatable():
                        attrs = {'expr': 'ALLOCATED(var%%%s)'%entity_name}
                        ifallocobj = part_append_genknode(subrobj, EXEC_PART, block_statements.IfThen, attrs=attrs)
                        if var.is_pointer():
                            attrs = {'expr': 'ASSOCIATED(var%%%s)'%entity_name}
                            ifassocobj = part_append_genknode(ifallocobj, EXEC_PART, block_statements.IfThen, attrs=attrs)
                    elif var.is_pointer():
                        attrs = {'expr': 'ASSOCIATED(var%%%s)'%entity_name}
                        ifassocobj = part_append_genknode(subrobj, EXEC_PART, block_statements.IfThen, attrs=attrs)

                    if ifassocobj: topobj = ifassocobj
                    elif ifallocobj: topobj = ifallocobj

                    attrs = {'variable': 'dtype_check_status%numTotal', 'sign': '=', 'expr': 'dtype_check_status%numTotal + 1'}
                    part_append_genknode(topobj, EXEC_PART, statements.Assignment, attrs=attrs)

                    if var.is_array():
                        dim_shape = ','.join(':'*var.rank)
                        get_size = ','.join(['SIZE(var%%%s,dim=%d)'%(entity_name, dim+1) for dim in range(var.rank)])

                        if stmt.is_derived():
                            indexes = [ 'idx%d'%(r+1) for r in range(var.rank)]
                            ename_indexes = [ '%s_%s'%(idx,entity_name) for idx in indexes ]

                            callname = None
                            for uname, req in stmt.unknowns.items():
                                if uname.firstpartname()==stmt.name and len(req.res_stmts)>0:
                                    callname = get_dtype_verifyname(req.res_stmts[0])
                                    break

                            attrs = {'type_spec': 'INTEGER', 'entity_decls': ename_indexes}
                            part_append_genknode(subrobj, DECL_PART, typedecl_statements.Integer, attrs=attrs)

                            attrs = {'designator': 'kgen_init_check', 'items': ['comp_check_status', 'rank=check_status%rank']}
                            part_append_genknode(topobj, EXEC_PART, statements.Call, attrs=attrs)

                            if callname:
                                pobj = topobj
                                doobjs = []
                                for d in range(var.rank):
                                    attrs = {'loopcontrol': ' idx%(d)d_%(e)s = LBOUND(var%%%(e)s,%(d)d), UBOUND(var%%%(e)s,%(d)d)'%{'e':entity_name, 'd':d+1}}
                                    doobj = part_append_genknode(pobj, EXEC_PART, block_statements.Do, attrs=attrs)
                                    doobjs.append(doobj)
                                    pobj = doobj 

                                # call verify subr
                                attrs = {'designator': callname, 'items': ['trim(adjustl(varname))//"%%%s"'%var.name, 'comp_check_status', \
                                    'var%%%s(%s)'%(entity_name, ','.join(ename_indexes)), 'kgenref_var%%%s(%s)'%(entity_name, ','.join(ename_indexes))]}
                                part_append_genknode(doobjs[-1], EXEC_PART, statements.Call, attrs=attrs)

                                attrs = {'expr': 'comp_check_status%numTotal == comp_check_status%numIdentical'}
                                ifidobj = part_append_genknode(topobj, EXEC_PART, block_statements.IfThen, attrs=attrs)

                                attrs = {'variable': 'dtype_check_status%numIdentical', 'sign': '=', 'expr': 'dtype_check_status%numIdentical + 1'}
                                part_append_genknode(ifidobj, EXEC_PART, statements.Assignment, attrs=attrs)

                                attrs = {'expr': 'kgen_verboseLevel > 2'}
                                ifvlobj = part_append_genknode(ifidobj, EXEC_PART, block_statements.IfThen, attrs=attrs)

                                attrs = {'expr': 'check_status%rank == 0'}
                                ifrank = part_append_genknode(ifvlobj, EXEC_PART, block_statements.IfThen, attrs=attrs)

                                attrs = {'items': ['trim(adjustl(varname))//"%%%s"'%entity_name,'" is IDENTICAL."']}
                                part_append_genknode(ifrank, EXEC_PART, statements.Write, attrs=attrs)

                                attrs = {'variable': 'check_result', 'sign': '=', 'expr': 'CHECK_IDENTICAL'}
                                part_append_genknode(ifidobj, EXEC_PART, statements.Assignment, attrs=attrs)

                                attrs = {'expr': 'comp_check_status%numOutTol > 0'}
                                part_append_genknode(ifidobj, EXEC_PART, block_statements.ElseIf, attrs=attrs)

                                attrs = {'variable': 'dtype_check_status%numOutTol', 'sign': '=', 'expr': 'dtype_check_status%numOutTol + 1'}
                                part_append_genknode(ifidobj, EXEC_PART, statements.Assignment, attrs=attrs)

                                attrs = {'expr': 'kgen_verboseLevel > 1'}
                                ifvlobj = part_append_genknode(ifidobj, EXEC_PART, block_statements.IfThen, attrs=attrs)

                                attrs = {'expr': 'check_status%rank == 0'}
                                ifrank = part_append_genknode(ifvlobj, EXEC_PART, block_statements.IfThen, attrs=attrs)

                                attrs = {'items': ['trim(adjustl(varname))','"%%%s is NOT IDENTICAL(out of tolerance)."'%entity_name]}
                                part_append_genknode(ifrank, EXEC_PART, statements.Write, attrs=attrs)

                                attrs = {'variable': 'check_result', 'sign': '=', 'expr': 'CHECK_OUT_TOL'}
                                part_append_genknode(ifidobj, EXEC_PART, statements.Assignment, attrs=attrs)

                                attrs = {'expr': 'comp_check_status%numInTol > 0'}
                                part_append_genknode(ifidobj, EXEC_PART, block_statements.ElseIf, attrs=attrs)

                                attrs = {'variable': 'dtype_check_status%numInTol', 'sign': '=', 'expr': 'dtype_check_status%numInTol + 1'}
                                part_append_genknode(ifidobj, EXEC_PART, statements.Assignment, attrs=attrs)

                                attrs = {'expr': 'kgen_verboseLevel > 1'}
                                ifvlobj = part_append_genknode(ifidobj, EXEC_PART, block_statements.IfThen, attrs=attrs)

                                attrs = {'expr': 'check_status%rank == 0'}
                                ifrank = part_append_genknode(ifvlobj, EXEC_PART, block_statements.IfThen, attrs=attrs)

                                attrs = {'items': ['trim(adjustl(varname))','"%%%s is NOT IDENTICAL(within tolerance)."'%entity_name]}
                                part_append_genknode(ifrank, EXEC_PART, statements.Write, attrs=attrs)

                                attrs = {'variable': 'check_result', 'sign': '=', 'expr': 'CHECK_IN_TOL'}
                                part_append_genknode(ifidobj, EXEC_PART, statements.Assignment, attrs=attrs)
                        else: # intrinsic type

                            attrs = {'type_spec': 'INTEGER', 'entity_decls': ['n_%s'%entity_name]}
                            part_append_genknode(subrobj, DECL_PART, typedecl_statements.Integer, attrs=attrs)

                            if stmt.name=='logical':
                                attrs = {'expr': 'ALL(var%%%s .EQV. kgenref_var%%%s)'%(entity_name, entity_name)}
                            else:
                                attrs = {'expr': 'ALL(var%%%s == kgenref_var%%%s)'%(entity_name, entity_name)}

                            ifidobj = part_append_genknode(topobj, EXEC_PART, block_statements.IfThen, attrs=attrs)

                            attrs = {'variable': 'dtype_check_status%numIdentical', 'sign': '=', 'expr': 'dtype_check_status%numIdentical + 1'}
                            part_append_genknode(ifidobj, EXEC_PART, statements.Assignment, attrs=attrs)

                            attrs = {'expr': 'kgen_verboseLevel > 2'}
                            ifvlobj = part_append_genknode(ifidobj, EXEC_PART, block_statements.IfThen, attrs=attrs)

                            attrs = {'expr': 'check_status%rank == 0'}
                            ifrank = part_append_genknode(ifvlobj, EXEC_PART, block_statements.IfThen, attrs=attrs)

                            attrs = {'items': ['trim(adjustl(varname))','"%%%s is IDENTICAL."'%entity_name]}
                            part_append_genknode(ifrank, EXEC_PART, statements.Write, attrs=attrs)

                            attrs = {'variable': 'check_result', 'sign': '=', 'expr': 'CHECK_IDENTICAL'}
                            part_append_genknode(ifidobj, EXEC_PART, statements.Assignment, attrs=attrs)

                            part_append_genknode(ifidobj, EXEC_PART, statements.Else)

                            if stmt.is_numeric():

                                # typececls
                                attrs = {'type_spec': stmt.name, 'selector':stmt.selector, 'entity_decls': \
                                    ['nrmsdiff_%s'%entity_name, 'rmsdiff_%s'%entity_name ]}
                                part_append_genknode(subrobj, DECL_PART, stmt.__class__, attrs=attrs)

                                attrs = {'type_spec': stmt.name, 'attrspec': ['ALLOCATABLE'], 'selector':stmt.selector, 'entity_decls': \
                                    ['buf1_%s(%s)'%(entity_name, dim_shape),'buf2_%s(%s)'%(entity_name, dim_shape)]}
                                part_append_genknode(subrobj, DECL_PART, stmt.__class__, attrs=attrs)

                                attrs = {'items': ['buf1_%s(%s)'%(entity_name, get_size)]}
                                part_append_genknode(ifidobj, EXEC_PART, statements.Allocate, attrs=attrs)

                                attrs = {'items': ['buf2_%s(%s)'%(entity_name, get_size)]}
                                part_append_genknode(ifidobj, EXEC_PART, statements.Allocate, attrs=attrs)

                                attrs = {'variable': 'n_%s'%entity_name, 'sign': '=', 'expr': 'COUNT(var%%%s /= kgenref_var%%%s)'%(entity_name, entity_name)}
                                part_append_genknode(ifidobj, EXEC_PART, statements.Assignment, attrs=attrs)

                                attrs = {'expr': 'ABS(kgenref_var%%%s) > kgen_minvalue'%entity_name}
                                whereobj = part_append_genknode(ifidobj, EXEC_PART, block_statements.Where, attrs=attrs)

                                attrs = {'variable': 'buf1_%s'%entity_name, 'sign': '=', \
                                    'expr': '((var%%%(e)s-kgenref_var%%%(e)s)/kgenref_var%%%(e)s)**2'%{'e':entity_name}}
                                part_append_genknode(whereobj, EXEC_PART, statements.Assignment, attrs=attrs)

                                attrs = {'variable': 'buf2_%s'%entity_name, 'sign': '=', 'expr': '(var%%%s-kgenref_var%%%s)**2'%(entity_name, entity_name)}
                                part_append_genknode(whereobj, EXEC_PART, statements.Assignment, attrs=attrs)

                                part_append_genknode(whereobj, EXEC_PART, statements.ElseWhere)

                                attrs = {'variable': 'buf1_%s'%entity_name, 'sign': '=', 'expr': '(var%%%s-kgenref_var%%%s)**2'%(entity_name, entity_name)}
                                part_append_genknode(whereobj, EXEC_PART, statements.Assignment, attrs=attrs)

                                attrs = {'variable': 'buf2_%s'%entity_name, 'sign': '=', 'expr': 'buf1_%s'%entity_name}
                                part_append_genknode(whereobj, EXEC_PART, statements.Assignment, attrs=attrs)

                                attrs = {'variable': 'nrmsdiff_%s'%entity_name, 'sign': '=', 'expr': 'SQRT(SUM(buf1_%s)/REAL(n_%s))'%(entity_name, entity_name)}
                                part_append_genknode(ifidobj, EXEC_PART, statements.Assignment, attrs=attrs)

                                attrs = {'variable': 'rmsdiff_%s'%entity_name, 'sign': '=', 'expr': 'SQRT(SUM(buf2_%s)/REAL(n_%s))'%(entity_name, entity_name)}
                                part_append_genknode(ifidobj, EXEC_PART, statements.Assignment, attrs=attrs)

                                #attrs = {'expr': 'nrmsdiff_%s > kgen_tolerance'%entity_name}
                                attrs = {'expr': 'rmsdiff_%s > kgen_tolerance'%entity_name}
                                ifvobj = part_append_genknode(ifidobj, EXEC_PART, block_statements.IfThen, attrs=attrs)

                                attrs = {'variable': 'dtype_check_status%numOutTol', 'sign': '=', 'expr': 'dtype_check_status%numOutTol + 1'}
                                part_append_genknode(ifvobj, EXEC_PART, statements.Assignment, attrs=attrs)

                                attrs = {'expr': 'kgen_verboseLevel > 1'}
                                ifvlobj = part_append_genknode(ifvobj, EXEC_PART, block_statements.IfThen, attrs=attrs)

                                attrs = {'expr': 'check_status%rank == 0'}
                                ifrank = part_append_genknode(ifvlobj, EXEC_PART, block_statements.IfThen, attrs=attrs)

                                attrs = {'items': ['trim(adjustl(varname))','"%%%s is NOT IDENTICAL(out of tolerance)."'%entity_name]}
                                part_append_genknode(ifrank, EXEC_PART, statements.Write, attrs=attrs)

                                attrs = {'variable': 'check_result', 'sign': '=', 'expr': 'CHECK_OUT_TOL'}
                                part_append_genknode(ifvobj, EXEC_PART, statements.Assignment, attrs=attrs)

                                part_append_genknode(ifvobj, EXEC_PART, statements.Else)

                                attrs = {'variable': 'dtype_check_status%numInTol', 'sign': '=', 'expr': 'dtype_check_status%numInTol + 1'}
                                part_append_genknode(ifvobj, EXEC_PART, statements.Assignment, attrs=attrs)

                                attrs = {'expr': 'kgen_verboseLevel > 1'}
                                ifvlobj = part_append_genknode(ifvobj, EXEC_PART, block_statements.IfThen, attrs=attrs)

                                attrs = {'expr': 'check_status%rank == 0'}
                                ifrank = part_append_genknode(ifvlobj, EXEC_PART, block_statements.IfThen, attrs=attrs)

                                attrs = {'items': ['trim(adjustl(varname))','"%%%s is NOT IDENTICAL(within tolerance)."'%entity_name]}
                                part_append_genknode(ifrank, EXEC_PART, statements.Write, attrs=attrs)

                                attrs = {'variable': 'check_result', 'sign': '=', 'expr': 'CHECK_IN_TOL'}
                                part_append_genknode(ifvobj, EXEC_PART, statements.Assignment, attrs=attrs)
                            else: # not numerical

                                if stmt.name=='logical':
                                    attrs = {'variable': 'n_%s'%entity_name, 'sign': '=', 'expr': 'COUNT(var%%%s .NEQV. kgenref_var%%%s)'%(entity_name, entity_name)}
                                else:
                                    attrs = {'variable': 'n_%s'%entity_name, 'sign': '=', 'expr': 'COUNT(var%%%s /= kgenref_var%%%s)'%(entity_name, entity_name)}

                                part_append_genknode(ifidobj, EXEC_PART, statements.Assignment, attrs=attrs)

                                attrs = {'variable': 'dtype_check_status%numOutTol', 'sign': '=', 'expr': 'dtype_check_status%numOutTol + 1'}
                                part_append_genknode(ifidobj, EXEC_PART, statements.Assignment, attrs=attrs)

                                attrs = {'expr': 'kgen_verboseLevel > 1'}
                                ifvlobj = part_append_genknode(ifidobj, EXEC_PART, block_statements.IfThen, attrs=attrs)

                                attrs = {'expr': 'check_status%rank == 0'}
                                ifrank = part_append_genknode(ifvlobj, EXEC_PART, block_statements.IfThen, attrs=attrs)

                                attrs = {'items': ['trim(adjustl(varname))','"%%%s is NOT IDENTICAL(out of tolerance)."'%entity_name]}
                                part_append_genknode(ifrank, EXEC_PART, statements.Write, attrs=attrs)

                                attrs = {'variable': 'check_result', 'sign': '=', 'expr': 'CHECK_OUT_TOL'}
                                part_append_genknode(ifidobj, EXEC_PART, statements.Assignment, attrs=attrs)
                    else: # scalar
                        if stmt.is_derived():

                            callname = None
                            for uname, req in stmt.unknowns.items():
                                if uname.firstpartname()==stmt.name and len(req.res_stmts)>0:
                                    callname = get_dtype_verifyname(req.res_stmts[0])
                                    break
                            #import pdb; pdb.set_trace()
                            if callname:

                                attrs = {'designator': 'kgen_init_check', 'items': ['comp_check_status', 'rank=check_status%rank']}
                                part_append_genknode(topobj, EXEC_PART, statements.Call, attrs=attrs)

                                attrs = {'designator': callname, 'items': ['"%s"'%entity_name, 'comp_check_status', 'var%%%s'%entity_name, 'kgenref_var%%%s'%entity_name]}
                                part_append_genknode(topobj, EXEC_PART, statements.Call, attrs=attrs)

                                attrs = {'expr': 'comp_check_status%numTotal == comp_check_status%numIdentical'}
                                ifidobj = part_append_genknode(topobj, EXEC_PART, block_statements.IfThen, attrs=attrs)

                                attrs = {'variable': 'dtype_check_status%numIdentical', 'sign': '=', 'expr': 'dtype_check_status%numIdentical + 1'}
                                part_append_genknode(ifidobj, EXEC_PART, statements.Assignment, attrs=attrs)

                                attrs = {'expr': 'kgen_verboseLevel > 2'}
                                ifvlobj = part_append_genknode(ifidobj, EXEC_PART, block_statements.IfThen, attrs=attrs)

                                attrs = {'expr': 'check_status%rank == 0'}
                                ifrank = part_append_genknode(ifvlobj, EXEC_PART, block_statements.IfThen, attrs=attrs)

                                attrs = {'items': ['trim(adjustl(varname))//"%%%s"'%entity_name,'" is IDENTICAL."']}
                                part_append_genknode(ifrank, EXEC_PART, statements.Write, attrs=attrs)

                                attrs = {'variable': 'check_result', 'sign': '=', 'expr': 'CHECK_IDENTICAL'}
                                part_append_genknode(ifidobj, EXEC_PART, statements.Assignment, attrs=attrs)

                                attrs = {'expr': 'comp_check_status%numOutTol > 0'}
                                part_append_genknode(ifidobj, EXEC_PART, block_statements.ElseIf, attrs=attrs)

                                attrs = {'variable': 'dtype_check_status%numOutTol', 'sign': '=', 'expr': 'dtype_check_status%numOutTol + 1'}
                                part_append_genknode(ifidobj, EXEC_PART, statements.Assignment, attrs=attrs)

                                attrs = {'expr': 'kgen_verboseLevel > 1'}
                                ifvlobj = part_append_genknode(ifidobj, EXEC_PART, block_statements.IfThen, attrs=attrs)

                                attrs = {'expr': 'check_status%rank == 0'}
                                ifrank = part_append_genknode(ifvlobj, EXEC_PART, block_statements.IfThen, attrs=attrs)

                                attrs = {'items': ['trim(adjustl(varname))','"%%%s is NOT IDENTICAL(out of tolerance)."'%entity_name]}
                                part_append_genknode(ifrank, EXEC_PART, statements.Write, attrs=attrs)

                                attrs = {'variable': 'check_result', 'sign': '=', 'expr': 'CHECK_OUT_TOL'}
                                part_append_genknode(ifidobj, EXEC_PART, statements.Assignment, attrs=attrs)

                                attrs = {'expr': 'comp_check_status%numInTol > 0'}
                                part_append_genknode(ifidobj, EXEC_PART, block_statements.ElseIf, attrs=attrs)

                                attrs = {'variable': 'dtype_check_status%numInTol', 'sign': '=', 'expr': 'dtype_check_status%numInTol + 1'}
                                part_append_genknode(ifidobj, EXEC_PART, statements.Assignment, attrs=attrs)

                                attrs = {'expr': 'kgen_verboseLevel > 1'}
                                ifvlobj = part_append_genknode(ifidobj, EXEC_PART, block_statements.IfThen, attrs=attrs)

                                attrs = {'expr': 'check_status%rank == 0'}
                                ifrank = part_append_genknode(ifvlobj, EXEC_PART, block_statements.IfThen, attrs=attrs)

                                attrs = {'items': ['trim(adjustl(varname))','"%%%s is NOT IDENTICAL(within tolerance)."'%entity_name]}
                                part_append_genknode(ifrank, EXEC_PART, statements.Write, attrs=attrs)

                                attrs = {'variable': 'check_result', 'sign': '=', 'expr': 'CHECK_IN_TOL'}
                                part_append_genknode(ifidobj, EXEC_PART, statements.Assignment, attrs=attrs)

                        else:
                            # diff
                            attrs = {'type_spec': stmt.name, 'selector':stmt.selector, 'entity_decls': ['diff_%s'%entity_name]}
                            part_append_genknode(subrobj, DECL_PART, stmt.__class__, attrs=attrs)

                            if stmt.name=='logical':
                                attrs = {'expr': 'var%%%s .EQV. kgenref_var%%%s'%(entity_name, entity_name)}
                            else:
                                attrs = {'expr': 'var%%%s == kgenref_var%%%s'%(entity_name, entity_name)}
                            ifidobj = part_append_genknode(topobj, EXEC_PART, block_statements.IfThen, attrs=attrs)

                            attrs = {'variable': 'dtype_check_status%numIdentical', 'sign': '=', 'expr': 'dtype_check_status%numIdentical + 1'}
                            part_append_genknode(ifidobj, EXEC_PART, statements.Assignment, attrs=attrs)

                            attrs = {'expr': 'kgen_verboseLevel > 2'}
                            ifvlobj = part_append_genknode(ifidobj, EXEC_PART, block_statements.IfThen, attrs=attrs)

                            attrs = {'expr': 'check_status%rank == 0'}
                            ifrank = part_append_genknode(ifvlobj, EXEC_PART, block_statements.IfThen, attrs=attrs)

                            attrs = {'items': ['trim(adjustl(varname))','"%%%s is IDENTICAL."'%entity_name]}
                            part_append_genknode(ifrank, EXEC_PART, statements.Write, attrs=attrs)

                            attrs = {'variable': 'check_result', 'sign': '=', 'expr': 'CHECK_IDENTICAL'}
                            part_append_genknode(ifidobj, EXEC_PART, statements.Assignment, attrs=attrs)

                            part_append_genknode(ifidobj, EXEC_PART, statements.Else)

                            if stmt.is_numeric():

                                attrs = {'variable': 'diff_%s'%entity_name, 'sign': '=', 'expr': 'ABS(var%%%s - kgenref_var%%%s)'%(entity_name, entity_name)}
                                part_append_genknode(ifidobj, EXEC_PART, statements.Assignment, attrs=attrs)

                                attrs = {'expr': 'diff_%s <= kgen_tolerance'%entity_name}
                                ifvobj = part_append_genknode(ifidobj, EXEC_PART, block_statements.IfThen, attrs=attrs)

                                attrs = {'variable': 'dtype_check_status%numInTol', 'sign': '=', 'expr': 'dtype_check_status%numInTol + 1'}
                                part_append_genknode(ifvobj, EXEC_PART, statements.Assignment, attrs=attrs)

                                attrs = {'expr': 'kgen_verboseLevel > 1'}
                                ifvlobj = part_append_genknode(ifvobj, EXEC_PART, block_statements.IfThen, attrs=attrs)

                                attrs = {'expr': 'check_status%rank == 0'}
                                ifrank = part_append_genknode(ifvlobj, EXEC_PART, block_statements.IfThen, attrs=attrs)

                                attrs = {'items': ['trim(adjustl(varname))','"%%%s is NOT IDENTICAL(within tolerance)."'%entity_name]}
                                part_append_genknode(ifrank, EXEC_PART, statements.Write, attrs=attrs)

                                attrs = {'variable': 'check_result', 'sign': '=', 'expr': 'CHECK_IN_TOL'}
                                part_append_genknode(ifvobj, EXEC_PART, statements.Assignment, attrs=attrs)

                                part_append_genknode(ifvobj, EXEC_PART, statements.Else)

                                attrs = {'variable': 'dtype_check_status%numOutTol', 'sign': '=', 'expr': 'dtype_check_status%numOutTol + 1'}
                                part_append_genknode(ifvobj, EXEC_PART, statements.Assignment, attrs=attrs)

                                attrs = {'expr': 'kgen_verboseLevel > 1'}
                                ifvlobj = part_append_genknode(ifvobj, EXEC_PART, block_statements.IfThen, attrs=attrs)

                                attrs = {'expr': 'check_status%rank == 0'}
                                ifrank = part_append_genknode(ifvlobj, EXEC_PART, block_statements.IfThen, attrs=attrs)

                                attrs = {'items': ['trim(adjustl(varname))','"%%%s is NOT IDENTICAL(out of tolerance)."'%entity_name]}
                                part_append_genknode(ifrank, EXEC_PART, statements.Write, attrs=attrs)

                                attrs = {'variable': 'check_result', 'sign': '=', 'expr': 'CHECK_OUT_TOL'}
                                part_append_genknode(ifvobj, EXEC_PART, statements.Assignment, attrs=attrs)
                            else: # not numeric

                                attrs = {'variable': 'dtype_check_status%numOutTol', 'sign': '=', 'expr': 'dtype_check_status%numOutTol + 1'}
                                part_append_genknode(ifidobj, EXEC_PART, statements.Assignment, attrs=attrs)

                                attrs = {'expr': 'kgen_verboseLevel > 1'}
                                ifvlobj = part_append_genknode(ifidobj, EXEC_PART, block_statements.IfThen, attrs=attrs)

                                attrs = {'expr': 'check_status%rank == 0'}
                                ifrank = part_append_genknode(ifvlobj, EXEC_PART, block_statements.IfThen, attrs=attrs)

                                attrs = {'items': ['trim(adjustl(varname))','"%%%s is NOT IDENTICAL."'%entity_name]}
                                part_append_genknode(ifrank, EXEC_PART, statements.Write, attrs=attrs)

                                attrs = {'variable': 'check_result', 'sign': '=', 'expr': 'CHECK_OUT_TOL'}
                                part_append_genknode(ifidobj, EXEC_PART, statements.Assignment, attrs=attrs)

                    print_detail = print_dummy_detail
                    if var.is_array(): # array
                        if stmt.is_derived():
                            print_detail = print_dtypearr_detail
                            pass
                        else:
                            if stmt.is_numeric():
                                print_detail = print_numarr_detail
                            else:
                                pass
                    else:
                        if stmt.is_derived():
                            print_detail = print_dtype_detail
                        else:
                            if stmt.is_numeric():
                                print_detail = print_num_detail
                            else:
                                pass

                    attrs = {'expr': 'check_result == CHECK_IDENTICAL'}
                    ifchkobj = part_append_genknode(topobj, EXEC_PART, block_statements.IfThen, attrs=attrs)

                    part_append_genknode(ifchkobj, EXEC_PART, statements.Continue)

#                    attrs = {'expr': 'kgen_verboseLevel > 2'}
#                    iflevel3obj = part_append_genknode(ifchkobj, EXEC_PART, block_statements.IfThen, attrs=attrs)
#
#                    print_detail(iflevel3obj, entity_name)

                    attrs = {'expr': 'check_result == CHECK_OUT_TOL'}
                    part_append_genknode(ifchkobj, EXEC_PART, block_statements.ElseIf, attrs=attrs)

                    attrs = {'expr': 'kgen_verboseLevel > 2'}
                    iflevel0obj = part_append_genknode(ifchkobj, EXEC_PART, block_statements.IfThen, attrs=attrs)

                    attrs = {'expr': 'check_status%rank ==0'}
                    ifrank = part_append_genknode(iflevel0obj, EXEC_PART, block_statements.IfThen, attrs=attrs)

                    print_detail(ifrank, entity_name)

                    attrs = {'expr': 'check_result == CHECK_IN_TOL'}
                    part_append_genknode(ifchkobj, EXEC_PART, block_statements.ElseIf, attrs=attrs)

                    attrs = {'expr': 'kgen_verboseLevel > 2'}
                    iflevel1obj = part_append_genknode(ifchkobj, EXEC_PART, block_statements.IfThen, attrs=attrs)

                    attrs = {'expr': 'check_status%rank ==0'}
                    ifrank = part_append_genknode(iflevel1obj, EXEC_PART, block_statements.IfThen, attrs=attrs)

                    print_detail(ifrank, entity_name)

                part_append_comment(topobj, EXEC_PART, '')


            attrs = {'expr': 'dtype_check_status%numTotal == dtype_check_status%numIdentical'}
            ifidobj = part_append_genknode(subrobj, EXEC_PART, block_statements.IfThen, attrs=attrs)

            attrs = {'variable': 'check_status%numIdentical', 'sign': '=', 'expr': 'check_status%numIdentical + 1'}
            part_append_genknode(ifidobj, EXEC_PART, statements.Assignment, attrs=attrs)

            attrs = {'expr': 'dtype_check_status%numOutTol > 0'}
            part_append_genknode(ifidobj, EXEC_PART, block_statements.ElseIf, attrs=attrs)

            attrs = {'variable': 'check_status%numOutTol', 'sign': '=', 'expr': 'check_status%numOutTol + 1'}
            part_append_genknode(ifidobj, EXEC_PART, statements.Assignment, attrs=attrs)

            attrs = {'expr': 'dtype_check_status%numInTol > 0'}
            part_append_genknode(ifidobj, EXEC_PART, block_statements.ElseIf, attrs=attrs)

            attrs = {'variable': 'check_status%numInTol', 'sign': '=', 'expr': 'check_status%numInTol + 1'}
            part_append_genknode(ifidobj, EXEC_PART, statements.Assignment, attrs=attrs)

            # create public stmt
            if parent.kgen_match_class in [ block_statements.Program, block_statements.Module ]:
                attrs = {'items': [subrname]}
                part_append_genknode(parent, DECL_PART, statements.Public, attrs=attrs)




'''KGen utilities
'''

import sys
import logging
import subprocess
import locale

try:
    import configparser
except:
    import ConfigParser as configparser

##############################################
# KGName
##############################################

EXTERNAL_NAMELEVEL_SEPERATOR = ':'
INTERNAL_NAMELEVEL_SEPERATOR = '__kgen__' # lower-case only

def encode_NS(namepath):
    return namepath.replace(EXTERNAL_NAMELEVEL_SEPERATOR, INTERNAL_NAMELEVEL_SEPERATOR)

def decode_NS(namepath):
    return namepath.replace(INTERNAL_NAMELEVEL_SEPERATOR, EXTERNAL_NAMELEVEL_SEPERATOR)

class KGName(object):
    def __init__(self, name, node=None, stmt=None):
        from fortlab.resolver.Fortran2003 import Data_Ref
        if not name: raise ProgramException('Name can not be none or blank')
        if name[0].isdigit(): raise ProgramException('Name can not have digit as its first character')

        self.namepath = encode_NS(name).strip().lower() # lower case
        self.namelist = self.namepath.split(INTERNAL_NAMELEVEL_SEPERATOR)
        self.dataref = Data_Ref(self.namelist[-1])
        self.node = node
        self.stmt = stmt
        #self.rename = []

    def path(self):
        return decode_NS(self.namepath)

    def list(self):
        return self.namelist

    def dataref(self):
        return self.dataref

    def last(self):
        return self.namelist[-1]

    def first(self):
        return self.namelist[0]

    def firstpartname(self):
        from fortlab.resolver.Fortran2003 import Name
        if isinstance(self.dataref, Name):
            return self.dataref.string
        else:
            return self.dataref.items[0].string

    def __eq__(self, other):
        return self.namepath==other.namepath

    def __str__(self):
        raise Exception('KGName')

    def __hash__(self):
        return id(self)


def _get_namepath(stmt, external):
    if external:
        return EXTERNAL_NAMELEVEL_SEPERATOR.join([ a.name.lower() for a in stmt.ancestors() ])
    else:
        return INTERNAL_NAMELEVEL_SEPERATOR.join([ a.name.lower() for a in stmt.ancestors() ])

def _pack_namepath(stmt, lastname, external):
    if external:
        return '%s%s%s'%(_get_namepath(stmt, True), EXTERNAL_NAMELEVEL_SEPERATOR, lastname)
    else:
        return '%s%s%s'%(_get_namepath(stmt, False), INTERNAL_NAMELEVEL_SEPERATOR, lastname)

def pack_innamepath(stmt, name):
    return _pack_namepath(stmt, name, False)

def pack_exnamepath(stmt, name):
    return _pack_namepath(stmt, name, True)

def get_innamepath(stmt):
    return _get_namepath(stmt, False)

def get_exnamepath(stmt):
    return _get_namepath(stmt, True)

def match_namepath(pattern, namepath, internal=True):

#name -> name in the beginning and the end
#:name: -> name in any location
#name: -> name at the beginning
#:name -> name at the end
#name1:name2 -> two level name
#name1:name2: -> more than two level name starts with the two names
#:name1:name2 -> more than two level name ends with the two names
#:name1:name2: -> more than two level name that the two name locates in the middle
#eventually data slicing

    if not pattern or not namepath: return False

    if internal:
        split_pattern = pattern.split(INTERNAL_NAMELEVEL_SEPERATOR)
        split_namepath = namepath.split(INTERNAL_NAMELEVEL_SEPERATOR)
    else:
        split_pattern = pattern.split(EXTERNAL_NAMELEVEL_SEPERATOR)
        split_namepath = namepath.split(EXTERNAL_NAMELEVEL_SEPERATOR)

    p = list(split_pattern)

    leading_mark = False
    if len(p[0])==0:
        leading_mark = True
        p = p[1:]

    ending_mark = False
    if len(p[-1])==0:
        ending_mark = True
        p = p[:-1]
        if len(p)==0:
            raise UserException('Wrong namepath format: %s'%split_pattern)

    n = list(split_namepath)
    while len(p)>0 and len(n)>0:
        if p[0]==n[0]:
            p = p[1:]
            n = n[1:]
        elif leading_mark:
            n = n[1:]
        elif len(p[0])==0:
            leading_mark = True
            p = p[1:]
        else:
            return False

    if len(p)==0:
        if len(n)>0:
            if ending_mark: return True
            return False
        else:
            return True
    else:
        if len(n)>0:
            raise ProgramException('Incorrect namepath match: (%s, %s)'%(split_pattern, split_namepath))
        else:
            return False

#############################################################################
## EXCEPTION
#############################################################################

class KGException(Exception):
    pass

class UserException(KGException):
    pass

class ProgramException(KGException):
    pass

##############################################
# Functions
##############################################

def tounicode(text, enc=None):

    if enc is None:
        if sys.stdout.encoding:
            enc = sys.stdout.encoding

        else:
            enc = "utf-8"

    try:
        text = text.decode(enc)

    except:
        text = str(text)

    return text

def kgenexit(msg):
    print(msg)
    sys.exit(-1)

def run_shcmd(cmd, input=None, **kwargs):

    show_error_msg = None
    if 'show_error_msg' in kwargs:
        show_error_msg = kwargs['show_error_msg']
        del kwargs['show_error_msg']

    enc = locale.getpreferredencoding(False)
    if type(cmd) == type(u""):
        cmd = cmd.encode(enc)

    if input and type(input) == type(u""):
        input = input.encode(enc)

    proc = subprocess.Popen(cmd, stdin=subprocess.PIPE, \
        stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True, **kwargs)

    out, err = proc.communicate(input=input)

    if proc.returncode != 0 and show_error_msg:
        print('>> %s' % cmd)
        print('returned non-zero code from shell('+str(ret_code)+')\n OUTPUT: '+str(out)+'\n ERROR: '+str(err)+'\n')

    return out, err, proc.returncode

# traverse f2003 nodes
# traverse and func will return None if to continue processing
# traverse and func will return return code if to stop processing
# The return code will be forwarded to initial caller
# func will collect anything in bag during processing
def traverse(node, func, bag, subnode='items', prerun=True, depth=0):
    ret = None

    if prerun and func is not None:
        ret = func(node, bag, depth)
        if ret is not None: return ret

    if node and hasattr(node, subnode) and getattr(node, subnode) is not None:
        for child in getattr(node, subnode):
            ret = traverse(child, func, bag, subnode=subnode, prerun=prerun, depth=depth+1)

    if not prerun and func is not None:
        ret = func(node, bag, depth)
        if ret is not None: return ret

    return ret

def get_subtree(obj, tree, prefix='top', depth=0):
    tab = '    '
    postfix = ''
    if isinstance(obj, str): postfix = ' => ' + obj
    elif isinstance(obj, type): postfix = ' => ' + str(obj)
    elif obj.__class__.__name__=='Name': postfix = ' => ' + obj.string

    #tree += [ ( tab*depth + prefix + ': ' + str(obj.__class__) + postfix, depth ) ]
    if hasattr(obj, 'parent'):
        pcls = str(obj.parent.__class__)
    else:
        pcls = 'None'
    tree += [ ( tab*depth + prefix + ': ' + str(obj.__class__) + postfix + ': parent => ' + pcls , depth ) ]
    if hasattr(obj, 'items'):
        for item in obj.items:
            get_subtree(item, tree, prefix='item', depth=depth+1)

    if hasattr(obj, 'content'):
        for elem in obj.content:
            get_subtree(elem, tree, prefix='content', depth=depth+1)

def show_obj(obj):
    print('CLS: ', obj.__class__)
    print('STR: ', str(obj))
    print('DIR: ', dir(obj))

def show_tree(node, prevent_print=False):
    tree = []
    get_subtree(node, tree)
    lines = []
    for elem, depth in tree:
        line = '    '*depth + elem
        if not prevent_print:
            print(line)
        lines.append(line+'\n')
    return lines

def chunks(l, n):
    for i in range(0, len(l), n):
        yield l[i:i + n]

def remove_multiblanklines(text):
    MAXBLANKLINES = 3
    lines = text.split('\n')
    newlines = []
    count = 0
    for line in lines:
        if len(line)>0:
            newlines.append(line)
            count = 0
        else:
            count += 1
            if count < MAXBLANKLINES:
                newlines.append(line)

    return '\n'.join(newlines)

def dequote(s):
    s = s.strip()
    if len(s) == 0: return s
    if (s[0] == s[-1]) and s.startswith(("'", '"')):
        return s[1:-1]
    return s

#
#def get_exclude_actions(section_name, config, *args ):
#    if section_name=='namepath':
#        if len(args)<1: return []
#
#        if section_name in config["exclude"]:
#            options = config["exclude"][section_name]
#            for pattern, actions in options.iteritems():
#                if match_namepath(pattern, args[0]):
#                    return actions
#        return []
#    else:
#        raise UserException('Not supported section name in exclusion input file: %s'%section)
#
##############################################
# Logging
##############################################

# create logger with 'spam_application'
logger = logging.getLogger('kgen')
#logger.setLevel(logging.DEBUG)
logger.setLevel(logging.WARNING)

# create file handler which logs even debug messages
fh = logging.FileHandler('kgen.log', mode='w')
fh.setLevel(logging.DEBUG)

# create console handler with a higher log level
ch = logging.StreamHandler()
ch.setLevel(logging.INFO)

# create formatter and add it to the handlers
cf = logging.Formatter('%(message)s')
ff = logging.Formatter('%(levelname)-8s [%(filename)s:%(lineno)s] %(message)s')
fh.setFormatter(ff)
ch.setFormatter(cf)

# add the handlers to the logger
logger.addHandler(fh)
logger.addHandler(ch)



#############################################################################
## CONFIG
#############################################################################

class KgenConfigParser(configparser.RawConfigParser):
    def __init__(self, *args, **kwargs):
        configparser.RawConfigParser.__init__(self, *args, **kwargs)
        self.optionxform = str

    def _optname_colon_to_dot(self, line):
        newline = line.strip()

        if len(newline)>0:
            if newline[0]==';': # comment
                return line
            elif newline[0]=='[' and newline[-1]==']': # filepath
                return line.replace(':', INTERNAL_NAMELEVEL_SEPERATOR)
            else: # else
                pos = line.find('=')
                if pos>0:
                    return line[:pos].replace(':', INTERNAL_NAMELEVEL_SEPERATOR) + line[pos:]
                else:
                    raise UserException('KGEN requires an equal symbol at each option line')
        else:
            return line

    def read(self, filenames):
        try:
            from io import StringIO

            if isinstance(filenames, str):
                filenames = [filenames]

        except:
            from StringIO import StringIO

            if isinstance(filenames, basestring):
                filenames = [filenames]

        for filename in filenames:
            try:
                fp = open(filename)
            except IOError:
                continue

            lines = []
            for line in fp.readlines():
                lines.append(self._optname_colon_to_dot(line))
            fp.close()

            buf = StringIO(''.join(lines))
            self._read(buf, filename)
